
import unittest

from pymap.parsing import Params
from pymap.parsing.exceptions import NotParseable
from pymap.parsing.primitives import Nil, Number, Atom, String, QuotedString, \
    LiteralString, List
from pymap.parsing.state import ParsingState, ParsingInterrupt, \
    ExpectContinuation


class TestNil(unittest.TestCase):

    def test_parse(self):
        ret, buf = Nil.parse(b'  nil  ', Params())
        self.assertIsInstance(ret, Nil)
        self.assertIsNone(ret.value)
        self.assertEqual(b'  ', buf)

    def test_parse_failure(self):
        with self.assertRaises(NotParseable):
            Nil.parse(b'', Params())
        with self.assertRaises(NotParseable):
            Nil.parse(b'niltest', Params())

    def test_bytes(self):
        nil = Nil()
        self.assertEqual(b'NIL', bytes(nil))


class TestNumber(unittest.TestCase):

    def test_parse(self):
        ret, buf = Number.parse(b'  123  ', Params())
        self.assertIsInstance(ret, Number)
        self.assertEqual(123, ret.value)
        self.assertEqual(b'  ', buf)

    def test_parse_failure(self):
        with self.assertRaises(NotParseable):
            Number.parse(b'abc', Params())
        with self.assertRaises(NotParseable):
            Number.parse(b'123abc', Params())

    def test_bytes(self):
        nil = Number(456)
        self.assertEqual(b'456', bytes(nil))


class TestAtom(unittest.TestCase):

    def test_parse(self):
        ret, buf = Atom.parse(b'  AtoM asdf  ', Params())
        self.assertIsInstance(ret, Atom)
        self.assertEqual(b'AtoM', ret.value)
        self.assertEqual(b' asdf  ', buf)

    def test_parse_failure(self):
        with self.assertRaises(NotParseable):
            Atom.parse(b'{}', Params())

    def test_bytes(self):
        nil = Atom(b'TEST.STUFF:asdf')
        self.assertEqual(b'TEST.STUFF:asdf', bytes(nil))


class TestString(unittest.TestCase):

    def test_quoted_parse(self):
        ret, buf = String.parse(br'  "one\"two\\three"  ', Params())
        self.assertIsInstance(ret, QuotedString)
        self.assertEqual(br'one"two\three', ret.value)
        self.assertEqual(b'  ', buf)

    def test_quoted_parse_empty(self):
        ret, buf = String.parse(br'  ""  ', Params())
        self.assertIsInstance(ret, QuotedString)
        self.assertEqual(br'', ret.value)
        self.assertEqual(b'  ', buf)

    def test_quoted_parse_failure(self):
        with self.assertRaises(NotParseable):
            String.parse(b'test', Params())
        with self.assertRaises(NotParseable):
            String.parse(b'"one\r\ntwo"', Params())
        with self.assertRaises(NotParseable):
            String.parse(br'"one\ two"', Params())
        with self.assertRaises(NotParseable):
            String.parse(b'"test', Params())

    def test_quoted_bytes(self):
        qstring1 = QuotedString(b'one"two\\three')
        self.assertEqual(b'"one\\"two\\\\three"', bytes(qstring1))
        qstring2 = QuotedString(b'test', b'"asdf"')
        self.assertEqual(b'"asdf"', bytes(qstring2))

    def test_literal_parse(self):
        state = ParsingState(continuations=[b'test\x01abc'])
        ret, buf = String.parse(b'{5}\r\n', Params(state))
        self.assertIsInstance(ret, LiteralString)
        self.assertEqual(b'test\x01', ret.value)
        self.assertFalse(ret.binary)
        self.assertEqual(b'abc', buf)

    def test_literal_parse_empty(self):
        state = ParsingState(continuations=[b'abc'])
        ret, buf = String.parse(b'{0}\r\n', Params(state))
        self.assertIsInstance(ret, LiteralString)
        self.assertEqual(b'', ret.value)
        self.assertEqual(b'abc', buf)

    def test_literal_plus(self):
        ret, buf = String.parse(b'{5+}\r\ntest\x01abc', Params())
        self.assertIsInstance(ret, LiteralString)
        self.assertEqual(b'test\x01', ret.value)
        self.assertFalse(ret.binary)
        self.assertEqual(b'abc', buf)

    def test_literal_binary(self):
        state = ParsingState(continuations=[b'\x00\x01\02abc'])
        ret, buf = String.parse(b'~{3}\r\n', Params(state))
        self.assertIsInstance(ret, LiteralString)
        self.assertEqual(b'\x00\x01\x02', ret.value)
        self.assertTrue(ret.binary)
        self.assertEqual(b'abc', buf)

    def test_literal_plus_binary(self):
        ret, buf = String.parse(b'~{3+}\r\n\x00\x01\02abc', Params())
        self.assertIsInstance(ret, LiteralString)
        self.assertEqual(b'\x00\x01\x02', ret.value)
        self.assertTrue(ret.binary)
        self.assertEqual(b'abc', buf)

    def test_literal_parse_failure(self):
        with self.assertRaises(NotParseable):
            String.parse(b'{}\r\n', Params())
        with self.assertRaises(NotParseable):
            String.parse(b'{10}', Params())
        with self.assertRaises(NotParseable):
            String.parse(b'{10}\r\nabc', Params())
        with self.assertRaises(NotParseable):
            state = ParsingState(continuations=[b'a'*9])
            String.parse(b'{10}\r\n', Params(state))
        with self.assertRaises(NotParseable):
            String.parse(b'{10+}\r\n' + (b'a'*9), Params())
        with self.assertRaises(ParsingInterrupt) as raised1:
            String.parse(b'{10}\r\n', Params())
        self.assertIsInstance(raised1.exception.expected, ExpectContinuation)
        with self.assertRaises(NotParseable) as raised2:
            String.parse(b'{4097}\r\n', Params())
        self.assertEqual(b'[TOOBIG]', bytes(raised2.exception.code))

    def test_literal_bytes(self):
        qstring1 = LiteralString(b'one\r\ntwo')
        self.assertEqual(b'{8}\r\none\r\ntwo', bytes(qstring1))
        qstring2 = LiteralString(b'')
        self.assertEqual(b'{0}\r\n', bytes(qstring2))

    def test_build_binary(self):
        ret = String.build(b'\x00\x01', True)
        self.assertEqual(b'\x00\x01', ret.value)
        self.assertTrue(ret.binary)
        self.assertEqual(b'~{2}\r\n\x00\x01', bytes(ret))


class TestList(unittest.TestCase):

    def test_parse(self):
        ret, buf = List.parse(
            b'  (ONE 2 (NIL) "four" )  ',
            Params(list_expected=[Nil, Number, Atom, String, List]))
        self.assertIsInstance(ret, List)
        self.assertEqual(4, len(ret.value))
        self.assertEqual(b'  ', buf)
        self.assertIsInstance(ret.value[0], Atom)
        self.assertEqual(b'ONE', ret.value[0].value)
        self.assertIsInstance(ret.value[1], Number)
        self.assertEqual(2, ret.value[1].value)
        self.assertIsInstance(ret.value[2], List)
        self.assertIsNone(ret.value[2].value[0].value)
        self.assertIsInstance(ret.value[3], QuotedString)
        self.assertEqual(b'four', ret.value[3].value)

    def test_parse_empty(self):
        ret, buf = List.parse(br'  ()  ', Params())
        self.assertIsInstance(ret, List)
        self.assertEqual([], ret.value)
        self.assertEqual(b'  ', buf)

    def test_parse_failure(self):
        with self.assertRaises(NotParseable):
            List.parse(b'{}', Params())
        with self.assertRaises(NotParseable):
            List.parse(b'("one"TWO)', Params(list_expected=[Atom, String]))
        with self.assertRaises(NotParseable):
            List.parse(b'(123 abc 456)', Params(list_expected=[Number]))

    def test_bytes(self):
        ret = List([QuotedString(b'abc'), Number(123), List([Nil()])])
        self.assertEqual(b'("abc" 123 (NIL))', bytes(ret))




from __future__ import annotations

import os.path
import uuid
from argparse import ArgumentParser, Namespace
from collections.abc import Mapping, AsyncIterator
from contextlib import closing, asynccontextmanager, AsyncExitStack
from datetime import datetime, timezone
from secrets import token_bytes
from typing import Any, Optional, Final

from pkg_resources import resource_listdir, resource_stream
from pysasl.creds import AuthenticationCredentials
from pysasl.hashing import Cleartext

from pymap.config import BackendCapability, IMAPConfig
from pymap.exceptions import AuthorizationFailure, NotAllowedError, \
    UserNotFound
from pymap.health import HealthStatus
from pymap.interfaces.backend import BackendInterface
from pymap.interfaces.login import LoginInterface, IdentityInterface
from pymap.parsing.message import AppendMessage
from pymap.parsing.specials.flag import Flag, Recent
from pymap.token import AllTokens
from pymap.user import UserMetadata

from .filter import FilterSet
from .mailbox import Message, MailboxData, MailboxSet
from ..session import BaseSession

__all__ = ['DictBackend', 'Config']


class DictBackend(BackendInterface):
    """Defines a backend that uses an in-memory dictionary for example usage
    and integration testing.

    """

    def __init__(self, login: Login, config: Config) -> None:
        super().__init__()
        self._login = login
        self._config = config
        self._status = HealthStatus()

    @property
    def login(self) -> Login:
        return self._login

    @property
    def config(self) -> Config:
        return self._config

    @property
    def status(self) -> HealthStatus:
        return self._status

    @classmethod
    def add_subparser(cls, name: str, subparsers: Any) -> ArgumentParser:
        parser = subparsers.add_parser(name, help='in-memory backend')
        parser.add_argument('--demo-data', action='store_true',
                            help='load initial demo data')
        parser.add_argument('--demo-user', default='demouser',
                            metavar='VAL', help='demo user ID')
        parser.add_argument('--demo-password', default='demopass',
                            metavar='VAL', help='demo user password')
        return parser

    @classmethod
    async def init(cls, args: Namespace, **overrides: Any) \
            -> tuple[DictBackend, Config]:
        config = Config.from_args(args, **overrides)
        login = Login(config)
        return cls(login, config), config

    async def start(self, stack: AsyncExitStack) -> None:
        pass


class Config(IMAPConfig):
    """The config implementation for the dict backend."""

    def __init__(self, args: Namespace, *, demo_data: bool,
                 demo_user: str, demo_password: str,
                 demo_data_resource: str = __name__,
                 admin_key: bytes = None, **extra: Any) -> None:
        admin_key = admin_key or token_bytes()
        super().__init__(args, hash_context=Cleartext(), admin_key=admin_key,
                         **extra)
        self._demo_data = demo_data
        self._demo_user = demo_user
        self._demo_password = demo_password
        self._demo_data_resource = demo_data_resource
        self.set_cache: dict[str, tuple[MailboxSet, FilterSet]] = {}

    @property
    def backend_capability(self) -> BackendCapability:
        return BackendCapability(idle=True, object_id=True, multi_append=True)

    @property
    def demo_data(self) -> bool:
        """True if demo data should be loaded at startup."""
        return self._demo_data

    @property
    def demo_data_resource(self) -> str:
        """Resource path of demo data files."""
        return self._demo_data_resource

    @property
    def demo_user(self) -> str:
        """A login name that is valid at startup, which defaults to
        ``demouser``.

        """
        return self._demo_user

    @property
    def demo_password(self) -> str:
        """The password for the :attr:`.demo_user` login name, which defaults
        to ``demopass``.

        """
        return self._demo_password

    @classmethod
    def parse_args(cls, args: Namespace) -> Mapping[str, Any]:
        return {**super().parse_args(args),
                'demo_data': args.demo_data,
                'demo_user': args.demo_user,
                'demo_password': args.demo_password}


class Session(BaseSession[Message]):
    """The session implementation for the dict backend."""

    def __init__(self, owner: str, config: Config, mailbox_set: MailboxSet,
                 filter_set: FilterSet) -> None:
        super().__init__(owner)
        self._config = config
        self._mailbox_set = mailbox_set
        self._filter_set = filter_set

    @property
    def config(self) -> Config:
        return self._config

    @property
    def mailbox_set(self) -> MailboxSet:
        return self._mailbox_set

    @property
    def filter_set(self) -> FilterSet:
        return self._filter_set


class Login(LoginInterface):
    """The login implementation for the dict backend."""

    def __init__(self, config: Config) -> None:
        super().__init__()
        self.config = config
        self.users_dict = {config.demo_user: UserMetadata(
            config, password=config.demo_password)}
        self.tokens_dict: dict[str, tuple[str, bytes]] = {}
        self._tokens = AllTokens()

    @property
    def tokens(self) -> AllTokens:
        return self._tokens

    async def authenticate(self, credentials: AuthenticationCredentials) \
            -> Identity:
        authcid = credentials.authcid
        token_key: Optional[bytes] = None
        role: Optional[str] = None
        if credentials.authcid_type == 'login-token':
            if authcid in self.tokens_dict:
                authcid, token_key = self.tokens_dict[authcid]
        elif credentials.authcid_type == 'admin-token':
            authcid = credentials.identity
            role = 'admin'
        try:
            metadata = await Identity(authcid, self, None).get()
        except UserNotFound:
            metadata = UserMetadata(self.config)
        await metadata.check_password(credentials, token_key=token_key)
        role = role or metadata.role
        if role != 'admin' and authcid != credentials.identity:
            raise AuthorizationFailure()
        return Identity(credentials.identity, self, role)


class Identity(IdentityInterface):
    """The identity implementation for the dict backend."""

    def __init__(self, name: str, login: Login, role: Optional[str]) -> None:
        super().__init__()
        self.login: Final = login
        self.config: Final = login.config
        self._name = name
        self._role = role

    @property
    def name(self) -> str:
        return self._name

    async def new_token(self, *, expiration: datetime = None) -> Optional[str]:
        token_id = uuid.uuid4().hex
        token_key = token_bytes()
        self.login.tokens_dict[token_id] = (self.name, token_key)
        return self.login.tokens.get_login_token(
            token_id, token_key, expiration=expiration)

    @asynccontextmanager
    async def new_session(self) -> AsyncIterator[Session]:
        identity = self.name
        config = self.config
        _ = await self.get()
        mailbox_set, filter_set = config.set_cache.get(identity, (None, None))
        if not mailbox_set or not filter_set:
            mailbox_set = MailboxSet()
            filter_set = FilterSet()
            if config.demo_data and identity == config.demo_user:
                await self._load_demo(config.demo_data_resource,
                                      mailbox_set, filter_set)
            config.set_cache[identity] = (mailbox_set, filter_set)
        yield Session(identity, config, mailbox_set, filter_set)

    async def _load_demo(self, resource: str, mailbox_set: MailboxSet,
                         filter_set: FilterSet) -> None:
        inbox = await mailbox_set.get_mailbox('INBOX')
        await self._load_demo_mailbox(resource, 'INBOX', inbox)
        mbx_names = sorted(resource_listdir(resource, 'demo'))
        for name in mbx_names:
            if name == 'sieve':
                await self._load_demo_sieve(resource, name, filter_set)
            elif name != 'INBOX':
                await mailbox_set.add_mailbox(name)
                mbx = await mailbox_set.get_mailbox(name)
                await self._load_demo_mailbox(resource, name, mbx)

    async def _load_demo_sieve(self, resource: str, name: str,
                               filter_set: FilterSet) -> None:
        path = os.path.join('demo', name)
        with closing(resource_stream(resource, path)) as sieve_stream:
            sieve = sieve_stream.read()
        await filter_set.put('demo', sieve)
        await filter_set.set_active('demo')

    async def _load_demo_mailbox(self, resource: str, name: str,
                                 mbx: MailboxData) -> None:
        path = os.path.join('demo', name)
        msg_names = sorted(resource_listdir(resource, path))
        for msg_name in msg_names:
            if msg_name == '.readonly':
                mbx._readonly = True
                continue
            elif msg_name.startswith('.'):
                continue
            msg_path = os.path.join(path, msg_name)
            with closing(resource_stream(resource, msg_path)) as msg_stream:
                flags_line = msg_stream.readline()
                msg_timestamp = float(msg_stream.readline())
                msg_data = msg_stream.read()
            msg_dt = datetime.fromtimestamp(msg_timestamp, timezone.utc)
            msg_flags = {Flag(flag) for flag in flags_line.split()}
            if Recent in msg_flags:
                msg_flags.remove(Recent)
                msg_recent = True
            else:
                msg_recent = False
            append_msg = AppendMessage(msg_data, msg_dt, frozenset(msg_flags))
            await mbx.append(append_msg, recent=msg_recent)

    async def get(self) -> UserMetadata:
        data = self.login.users_dict.get(self.name)
        if data is None:
            raise UserNotFound(self.name)
        return data

    async def set(self, data: UserMetadata) -> None:
        if self._role != 'admin' and data.role:
            raise NotAllowedError('Cannot assign role.')
        self.login.users_dict[self.name] = data

    async def delete(self) -> None:
        try:
            del self.login.users_dict[self.name]
        except KeyError as exc:
            raise UserNotFound(self.name) from exc
        self.config.set_cache.pop(self.name, None)
        for token_id, (name, _) in list(self.login.tokens_dict.items()):
            if name == self.name:
                del self.login.tokens_dict[token_id]




from __future__ import annotations

from typing import Optional

from grpclib.server import Stream
from pymap.user import UserMetadata
from pymapadmin.grpc.admin_grpc import UserBase
from pymapadmin.grpc.admin_pb2 import \
    UserData, UserResponse, GetUserRequest, SetUserRequest, DeleteUserRequest

from . import BaseHandler

__all__ = ['UserHandlers']

_GetUserStream = Stream[GetUserRequest, UserResponse]
_SetUserStream = Stream[SetUserRequest, UserResponse]
_DeleteUserStream = Stream[DeleteUserRequest, UserResponse]


class UserHandlers(UserBase, BaseHandler):
    """The GRPC handlers, executed when an admin request is received. Each
    handler should receive a request, take action, and send the response.

    See Also:
        :class:`grpclib.server.Server`

    """

    async def GetUser(self, stream: _GetUserStream) -> None:
        """

        See ``pymap-admin get-user --help`` for more options.

        Args:
            stream (:class:`~grpclib.server.Stream`): The stream for the
                request and response.

        """
        request = await stream.recv_message()
        assert request is not None
        username: Optional[str] = None
        user_data: Optional[UserData] = None
        async with self.catch_errors('GetUser') as result, \
                self.login_as(stream.metadata, request.user) as identity:
            username = identity.name
            metadata = await identity.get()
            user_data = UserData(params=metadata.params)
            if metadata.password is not None:
                user_data.password = metadata.password
        resp = UserResponse(result=result, username=username,
                            data=user_data)
        await stream.send_message(resp)

    async def SetUser(self, stream: _SetUserStream) -> None:
        """

        See ``pymap-admin set-user --help`` for more options.

        Args:
            stream (:class:`~grpclib.server.Stream`): The stream for the
                request and response.

        """
        request = await stream.recv_message()
        assert request is not None
        async with self.catch_errors('SetUser') as result, \
                self.login_as(stream.metadata, request.user) as identity:
            user_data = request.data
            password = self.config.hash_context.hash(user_data.password) \
                if user_data.HasField('password') else None
            metadata = UserMetadata(self.config, password=password,
                                    **user_data.params)
            await identity.set(metadata)
        resp = UserResponse(result=result, username=request.user)
        await stream.send_message(resp)

    async def DeleteUser(self, stream: _DeleteUserStream) -> None:
        """

        See ``pymap-admin delete-user --help`` for more options.

        Args:
            stream (:class:`~grpclib.server.Stream`): The stream for the
                request and response.

        """
        request = await stream.recv_message()
        assert request is not None
        async with self.catch_errors('DeleteUser') as result, \
                self.login_as(stream.metadata, request.user) as identity:
            await identity.delete()
        resp = UserResponse(result=result, username=request.user)
        await stream.send_message(resp)




from __future__ import annotations

from collections.abc import Sequence
from typing import Optional

from pymap.mime import MessageContent
from pymap.parsing.message import AppendMessage
from sievelib.commands import Command

from .tests import SieveTest

__all__ = ['SieveRunner']


class StopRunning(Exception):
    pass


class SieveRunner:

    def __init__(self, commands: Sequence[Command]) -> None:
        super().__init__()
        self._commands = commands

    def get_actions(self, sender: str, recipient: str,
                    append_msg: AppendMessage) -> Sequence[Command]:
        actions: list[Command] = []
        try:
            content = MessageContent.parse(append_msg.literal)
            self._get_actions(actions, sender, recipient, append_msg, content)
        except StopRunning:
            pass
        return actions

    def _get_actions(self, actions: list[Command], sender: str, recipient: str,
                     append_msg: AppendMessage, content: MessageContent) \
            -> None:
        running_if: Optional[Command] = None
        for cmd in self._commands:
            cmd_type = cmd.get_type()
            if cmd_type == 'action':
                actions.append(cmd)
            elif cmd_type == 'control':
                if cmd.name == 'if':
                    running_if = cmd
                elif cmd.name in ('elsif', 'else'):
                    pass
                elif cmd.name == 'stop':
                    raise StopRunning()
                elif cmd.name == 'require':
                    pass
                else:
                    raise NotImplementedError(cmd.name)
            else:
                raise NotImplementedError(cmd_type)
            if running_if:
                if cmd.name in ('if', 'elsif'):
                    test = SieveTest.of(cmd.arguments['test'])
                    test_result = test.run(sender, recipient, append_msg,
                                           content)
                    if not test_result:
                        continue
                elif cmd.name == 'else':
                    pass
                else:
                    running_if = None
                    continue
                running_if = None
                runner = SieveRunner(cmd.children)
                runner._get_actions(actions, sender, recipient, append_msg,
                                    content)



"""Contains the context variables that are made available by the IMAP server to
backend implementations.

See Also:
    :mod:`contextvars`

"""

from __future__ import annotations

from contextlib import AsyncExitStack
from contextvars import ContextVar

from proxyprotocol.sock import SocketInfo

from .cluster import ClusterMetadata
from .concurrent import Subsystem
from .parsing.command import Command

__all__ = ['subsystem', 'current_command', 'socket_info', 'language_code',
           'connection_exit', 'cluster_metadata']

#: The :class:`~pymap.concurrent.Subsystem` for concurrency primitives.
subsystem: ContextVar[Subsystem] = ContextVar(
    'subsystem', default=Subsystem.for_asyncio())

#: The currently executing :class:`~pymap.parsing.command.Command`.
current_command: ContextVar[Command] = ContextVar('current_command')

#: :class:`~proxyprotocol.sock.SocketInfo` about the currently connected
#: client.
socket_info: ContextVar[SocketInfo] = ContextVar('socket_info')

#: The language code string, e.g. ``en``.
language_code: ContextVar[str] = ContextVar('language_code')

#: The active :class:`~contextlib.AsyncExitStack` that will be exited when the
#: connection closes.
connection_exit: ContextVar[AsyncExitStack] = ContextVar('connection_exit')

#: Manages local cluster metadata updates and access to remote status and
#: metadata.
cluster_metadata: ContextVar[ClusterMetadata] = ContextVar(
    'cluster_metdata', default=ClusterMetadata())




from __future__ import annotations

from abc import abstractmethod, ABCMeta
from collections.abc import Sequence, Mapping
from typing import Union, Final

from pymap.bytes import MaybeBytes, BytesFormat

__all__ = ['RedisKey', 'KeysGroup', 'GlobalKeys', 'CleanupKeys',
           'NamespaceKeys', 'ContentKeys', 'FilterKeys', 'MailboxKeys']

#: The version number of this key layout, to track backwards incompatible
#: changes.
DATA_VERSION: Final = 4

_Value = Union[int, MaybeBytes]


class RedisKey:
    """Defines complex redis keys using composite segments.

    Args:
        joiner: The object used to join segments.
        segments: The current list of key segments.
        named: A mapping of named segments.

    """

    __slots__ = ['joiner', 'segments', 'named']

    def __init__(self, joiner: BytesFormat, segments: Sequence[_Value],
                 named: Mapping[str, bytes]) -> None:
        super().__init__()
        self.joiner: Final = joiner
        self.segments: Final = segments
        self.named: Final = named

    def fork(self, segment: _Value, *, name: str = None) -> RedisKey:
        """Fork a new redis key adding a new segment.

        Args:
            segment: The new key segment to append.
            name: An optional name to assign to the value.

        """
        new_segments = list(self.segments)
        new_segments.append(segment)
        if name is not None:
            new_named = dict(self.named)
            new_named[name] = bytes(segment)
            return RedisKey(self.joiner, new_segments, new_named)
        else:
            return RedisKey(self.joiner, new_segments, self.named)

    def end(self, *segments: _Value) -> bytes:
        """Terminate the redis key producing a bytestring.

        Args:
            segments: Optional segments to add to the result.

        """
        return self.joiner.join(self.segments, segments)


class KeysGroup(metaclass=ABCMeta):
    """A group of redis keys that are related and contained in the same root
    key.

    Args:
        root: The root redis key.

    """

    __slots__ = ['root']

    def __init__(self, root: RedisKey) -> None:
        super().__init__()
        self.root = root

    @property
    @abstractmethod
    def keys(self) -> Sequence[bytes]:
        """All the keys in the group, not including nested key groups."""
        ...


class GlobalKeys(KeysGroup):
    """The key group that is considered global. New :attr:`DATA_VERSION` values
    must be compatible with these keys.

    Args:
        root: The root redis key.

    """

    __slots__ = ['cleanup_root', 'namespace_root', 'namespaces']

    def __init__(self, root: RedisKey) -> None:
        super().__init__(root)
        self.cleanup_root: Final = root.fork(b'cleanup')
        self.namespace_root: Final = root.fork(b'ns')
        self.namespaces: Final = root.end(b'ns')

    @property
    def keys(self) -> Sequence[bytes]:
        return [self.namespaces]


class CleanupKeys(KeysGroup):
    """The key group for cleaning up keys that are no longer active.

    Args:
        root: The root redis key.

    """

    __slots__ = ['namespaces', 'mailboxes', 'contents']

    def __init__(self, parent: GlobalKeys) -> None:
        root = parent.cleanup_root.fork(DATA_VERSION, name='version')
        super().__init__(root)
        self.namespaces: Final = root.end(b'ns')
        self.mailboxes: Final = root.end(b'mbx')
        self.contents: Final = root.end(b'content')

    @property
    def keys(self) -> Sequence[bytes]:
        return [self.namespaces, self.mailboxes, self.contents]


class NamespaceKeys(KeysGroup):
    """The key group for managing mailbox namespaces, groups of mailboxes that
    typically correspond to a login user.

    Args:
        root: The root redis key.
        namespace: The namespace ID.

    """

    __slots__ = ['mailbox_root', 'content_root', 'mailboxes', 'max_order',
                 'order', 'uid_validity', 'subscribed', 'content_refs',
                 'thread_keys', 'max_modseq']

    def __init__(self, parent: GlobalKeys, namespace: _Value) -> None:
        root = parent.namespace_root.fork(namespace, name='namespace')
        super().__init__(root)
        self.mailbox_root: Final = root.fork(b'mbx')
        self.content_root: Final = root.fork(b'content')
        self.mailboxes: Final = root.end(b'mailboxes')
        self.max_order: Final = root.end(b'max-order')
        self.order: Final = root.end(b'order')
        self.uid_validity: Final = root.end(b'uidv')
        self.subscribed: Final = root.end(b'subscribed')
        self.content_refs: Final = root.end(b'contentrefs')
        self.thread_keys: Final = root.end(b'threadkeys')
        self.max_modseq: Final = root.end(b'maxmodseq')

    @property
    def keys(self) -> Sequence[bytes]:
        return [self.mailboxes, self.max_order, self.order, self.uid_validity,
                self.subscribed, self.content_refs, self.thread_keys,
                self.max_modseq]


class ContentKeys(KeysGroup):
    """The key group for managing the content of messages.

    Args:
        root: The root redis key.
        email_id: The email object ID.

    """

    __slots__ = ['data']

    def __init__(self, parent: NamespaceKeys, email_id: _Value) -> None:
        root = parent.content_root.fork(email_id, name='email_id')
        super().__init__(root)
        self.data: Final = root.end()

    @property
    def keys(self) -> Sequence[bytes]:
        return [self.data]


class FilterKeys(KeysGroup):
    """The key group for managing filters.

    Args:
        root: The root redis key.

    """

    __slots__ = ['data', 'names']

    def __init__(self, parent: NamespaceKeys) -> None:
        root = parent.root.fork(b'sieve')
        super().__init__(root)
        self.data: Final = root.end(b'data')
        self.names: Final = root.end(b'names')

    @property
    def keys(self) -> Sequence[bytes]:
        return [self.data, self.names]


class MailboxKeys(KeysGroup):
    """The key group for managing a single mailbox.

    Args:
        root: The root redis key.
        mailbox_id: The mailbox object ID.

    """

    __slots__ = ['max_uid', 'uids', 'seq', 'content', 'changes', 'recent',
                 'deleted', 'unseen']

    def __init__(self, parent: NamespaceKeys, mailbox_id: _Value) -> None:
        root = parent.mailbox_root.fork(mailbox_id, name='mailbox_id')
        super().__init__(root)
        self.max_uid: Final = root.end(b'max-uid')
        self.uids: Final = root.end(b'uids')
        self.seq: Final = root.end(b'seq')
        self.content: Final = root.end(b'content')
        self.changes: Final = root.end(b'changes')
        self.recent: Final = root.end(b'recent')
        self.deleted: Final = root.end(b'deleted')
        self.unseen: Final = root.end(b'unseen')

    @property
    def keys(self) -> Sequence[bytes]:
        return [self.max_uid, self.uids, self.seq, self.content, self.changes,
                self.recent, self.deleted, self.unseen]




from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import Final

from aioredis import Redis

from . import ScriptBase
from ..keys import CleanupKeys, NamespaceKeys

__all__ = ['NamespaceScripts']


class NamespaceScripts:

    def __init__(self) -> None:
        super().__init__()
        self.list: Final = MailboxList()
        self.get: Final = MailboxGet()
        self.add: Final = MailboxAdd()
        self.delete: Final = MailboxDelete()


class MailboxList(ScriptBase[Sequence[bytes]]):

    def __init__(self) -> None:
        super().__init__('mailbox_list')

    def _convert(self, ret: tuple[Mapping[bytes, bytes], Sequence[bytes]]) \
            -> Sequence[bytes]:
        mailboxes, mbx_order = ret
        mailboxes_iter = iter(mailboxes)
        mailboxes_zip = zip(mailboxes_iter, mailboxes_iter)
        rev_mbx = {mbx_id: key for key, mbx_id in mailboxes_zip}
        return [rev_mbx[mbx_id] for mbx_id in mbx_order if mbx_id in rev_mbx]

    async def __call__(self, redis: Redis, ns_keys: NamespaceKeys) \
            -> Sequence[bytes]:
        keys = [ns_keys.mailboxes, ns_keys.order]
        return await self.eval(redis, keys, [])


class MailboxGet(ScriptBase[tuple[bytes, int]]):

    def __init__(self) -> None:
        super().__init__('mailbox_get')

    def _convert(self, ret: tuple[bytes, bytes]) -> tuple[bytes, int]:
        return (ret[0], int(ret[1]))

    async def __call__(self, redis: Redis, ns_keys: NamespaceKeys, *,
                       name: bytes) -> tuple[bytes, int]:
        keys = [ns_keys.mailboxes, ns_keys.uid_validity]
        return await self.eval(redis, keys, [name])


class MailboxAdd(ScriptBase[None]):

    def __init__(self) -> None:
        super().__init__('mailbox_add')

    async def __call__(self, redis: Redis, ns_keys: NamespaceKeys, *,
                       name: bytes, mailbox_id: bytes) -> None:
        keys = [ns_keys.mailboxes, ns_keys.order, ns_keys.max_order,
                ns_keys.uid_validity]
        return await self.eval(redis, keys, [name, mailbox_id])


class MailboxDelete(ScriptBase[None]):

    def __init__(self) -> None:
        super().__init__('mailbox_delete')

    async def __call__(self, redis: Redis,
                       ns_keys: NamespaceKeys, cl_keys: CleanupKeys, *,
                       name: bytes) -> None:
        keys = [ns_keys.mailboxes, ns_keys.order, cl_keys.mailboxes]
        return await self.eval(redis, keys, [
            name,
            ns_keys.root.named['namespace']])




from __future__ import annotations

from collections.abc import Iterable, Mapping, Sequence
from datetime import datetime
from typing import Optional

import msgpack
from aioredis import Redis, ResponseError, WatchError

from pymap.bytes import HashStream
from pymap.concurrent import Event
from pymap.exceptions import MailboxNotFound, MailboxConflict, TemporaryFailure
from pymap.flags import FlagOp
from pymap.interfaces.message import CachedMessage
from pymap.listtree import ListTree
from pymap.mailbox import MailboxSnapshot
from pymap.mime import MessageContent
from pymap.parsing.message import AppendMessage
from pymap.parsing.modutf7 import modutf7_encode, modutf7_decode
from pymap.parsing.specials import ObjectId, SequenceSet
from pymap.parsing.specials.flag import Flag
from pymap.selected import SelectedSet, SelectedMailbox
from pymap.threads import ThreadKey

from .keys import CleanupKeys, NamespaceKeys, ContentKeys, MailboxKeys
from .message import Message
from .scripts.mailbox import MailboxScripts
from .scripts.namespace import NamespaceScripts
from ..mailbox import MailboxDataInterface, MailboxSetInterface

__all__ = ['Message', 'MailboxData', 'MailboxSet']

_scripts = MailboxScripts()
_ns_scripts = NamespaceScripts()

_ChangesRaw = Sequence[tuple[bytes, Mapping[bytes, bytes]]]


class MailboxData(MailboxDataInterface[Message]):
    """Implementation of :class:`~pymap.backend.mailbox.MailboxDataInterface`
    for the redis backend.

    """

    def __init__(self, redis: Redis, mailbox_id: bytes, uid_validity: int,
                 keys: MailboxKeys, ns_keys: NamespaceKeys,
                 cl_keys: CleanupKeys) -> None:
        super().__init__()
        self._redis = redis
        self._mailbox_id = ObjectId(mailbox_id)
        self._uid_validity = uid_validity
        self._selected_set = SelectedSet()
        self._keys = keys
        self._ns_keys = ns_keys
        self._cl_keys = cl_keys

    @property
    def mailbox_id(self) -> ObjectId:
        return self._mailbox_id

    @property
    def readonly(self) -> bool:
        return False

    @property
    def uid_validity(self) -> int:
        return self._uid_validity

    @property
    def selected_set(self) -> SelectedSet:
        return self._selected_set

    def _get_msg(self, uid: int, msg_raw: bytes) -> Message:
        msg = msgpack.unpackb(msg_raw, raw=True)
        msg_flags = {Flag(flag) for flag in msg[b'flags']}
        msg_email_id = ObjectId.maybe(msg[b'email_id'])
        msg_thread_id = ObjectId.maybe(msg[b'thread_id'])
        msg_time = datetime.fromisoformat(msg[b'date'].decode('ascii'))
        return Message(uid, msg_time, msg_flags,
                       email_id=msg_email_id, thread_id=msg_thread_id,
                       redis=self._redis, ns_keys=self._ns_keys)

    async def update_selected(self, selected: SelectedMailbox, *,
                              wait_on: Event = None) -> SelectedMailbox:
        last_mod_seq = selected.mod_sequence
        if wait_on is not None:
            await self._wait_updates(selected, last_mod_seq)
        if last_mod_seq is None:
            await self._load_initial(selected)
        else:
            await self._load_updates(selected, last_mod_seq)
        return selected

    async def append(self, append_msg: AppendMessage, *,
                     recent: bool = False) -> Message:
        redis = self._redis
        keys = self._keys
        ns_keys = self._ns_keys
        when = append_msg.when or datetime.now()
        content = MessageContent.parse(append_msg.literal)
        content_hash = HashStream().digest(content)
        new_email_id = ObjectId.new_email_id(content_hash)
        ct_keys = ContentKeys(ns_keys, new_email_id)
        new_uid, email_id, thread_id = await _scripts.add(
            redis, ns_keys, ct_keys, keys,
            recent=recent, flags=[str(flag) for flag in append_msg.flag_set],
            date=when.isoformat().encode('ascii'),
            email_id=new_email_id.value,
            thread_id=ObjectId.random_thread_id().value,
            thread_keys=['\0'.join(thread_key) for thread_key in
                         ThreadKey.get_all(content.header)],
            message=append_msg.literal, message_json=content.json,
            header=bytes(content.header), header_json=content.header.json)
        return Message(new_uid, when, append_msg.flag_set,
                       email_id=ObjectId(email_id),
                       thread_id=ObjectId(thread_id),
                       redis=redis, ns_keys=ns_keys)

    async def copy(self, uid: int, destination: MailboxData, *,
                   recent: bool = False) -> Optional[int]:
        redis = self._redis
        keys = self._keys
        ns_keys = self._ns_keys
        dest_keys = destination._keys
        try:
            dest_uid = await _scripts.copy(redis, ns_keys, keys, dest_keys,
                                           source_uid=uid, recent=recent)
        except ResponseError as exc:
            if 'message not found' in str(exc):
                return None
            raise
        return dest_uid

    async def move(self, uid: int, destination: MailboxData, *,
                   recent: bool = False) -> Optional[int]:
        redis = self._redis
        keys = self._keys
        ns_keys = self._ns_keys
        dest_keys = destination._keys
        try:
            dest_uid = await _scripts.move(redis, ns_keys, keys, dest_keys,
                                           source_uid=uid, recent=recent)
        except ResponseError as exc:
            if 'message not found' in str(exc):
                return None
            raise
        return dest_uid

    async def get(self, uid: int, cached_msg: CachedMessage) -> Message:
        redis = self._redis
        keys = self._keys
        message_raw = await redis.hget(keys.uids, uid)
        if message_raw is None:
            if not isinstance(cached_msg, Message):
                raise TypeError(cached_msg)
            return Message.copy_expunged(cached_msg)
        return self._get_msg(uid, message_raw)

    async def update(self, uid: int, cached_msg: CachedMessage,
                     flag_set: frozenset[Flag], mode: FlagOp) -> Message:
        keys = self._keys
        ns_keys = self._ns_keys
        try:
            message_raw = await _scripts.update(
                self._redis, ns_keys, keys, uid=uid, mode=bytes(mode),
                flags=[str(flag) for flag in flag_set])
        except ResponseError as exc:
            if 'message not found' not in str(exc):
                raise
            if not isinstance(cached_msg, Message):
                raise TypeError(cached_msg)
            return Message.copy_expunged(cached_msg)
        return self._get_msg(uid, message_raw)

    async def delete(self, uids: Iterable[int]) -> None:
        keys = self._keys
        ns_keys = self._ns_keys
        uids = list(uids)
        if not uids:
            return
        await _scripts.delete(self._redis, ns_keys, keys, self._cl_keys,
                              uids=uids)

    async def claim_recent(self, selected: SelectedMailbox) -> None:
        keys = self._keys
        async with self._redis.pipeline() as multi:
            multi.smembers(keys.recent)
            multi.unlink(keys.recent)
            recent, _ = await multi.execute()
        for uid_bytes in recent:
            selected.session_flags.add_recent(int(uid_bytes))

    async def find_deleted(self, seq_set: SequenceSet,
                           selected: SelectedMailbox) -> Sequence[int]:
        if not seq_set.is_all:
            return await super().find_deleted(seq_set, selected)
        deleted = await self._redis.smembers(self._keys.deleted)
        return [int(uid) for uid in deleted]

    async def cleanup(self) -> None:
        pass

    async def snapshot(self) -> MailboxSnapshot:
        next_uid, num_exists, num_recent, num_unseen, first_unseen = \
            await _scripts.snapshot(self._redis, self._keys)
        return MailboxSnapshot(self.mailbox_id, self.readonly,
                               self.uid_validity, self.permanent_flags,
                               self.session_flags, num_exists, num_recent,
                               num_unseen, first_unseen, next_uid)

    def _get_mod_seq(self, changes: _ChangesRaw) -> Optional[bytes]:
        try:
            ret = changes[-1][0]
        except IndexError:
            return None
        else:
            left, right = ret.rsplit(b'-', 1)
            next_right = int(right) + 1
            return b'%b-%i' % (left, next_right)

    async def _load_initial(self, selected: SelectedMailbox) -> None:
        keys = self._keys
        async with self._redis.pipeline() as multi:
            multi.hgetall(keys.uids)
            multi.xrevrange(keys.changes, count=1)
            msg_raw_map, last_changes = await multi.execute()
        messages = [self._get_msg(int(uid), msg_raw)
                    for uid, msg_raw in msg_raw_map.items()]
        selected.mod_sequence = self._get_mod_seq(last_changes)
        selected.set_messages(messages)

    def _get_changes(self, changes: _ChangesRaw) \
            -> tuple[Sequence[Message], frozenset[int]]:
        expunged = frozenset(int(fields[b'uid']) for _, fields in changes
                             if fields[b'type'] == b'expunge')
        messages: list[Message] = []
        for _, fields in changes:
            uid = int(fields[b'uid'])
            if fields[b'type'] != b'fetch' or uid in expunged:
                continue
            msg = self._get_msg(uid, fields[b'message'])
            messages.append(msg)
        return messages, expunged

    async def _load_updates(self, selected: SelectedMailbox,
                            last_mod_seq: bytes) -> None:
        keys = self._keys
        async with self._redis.pipeline() as multi:
            multi.xlen(keys.changes)
            multi.xrange(keys.changes, min=last_mod_seq)
            multi.xrevrange(keys.changes, count=1)
            changes_len, changes, last_changes = await multi.execute()
        if len(changes) == changes_len:
            return await self._load_initial(selected)
        messages, expunged = self._get_changes(changes)
        selected.mod_sequence = self._get_mod_seq(last_changes)
        selected.add_updates(messages, expunged)

    async def _wait_updates(self, selected: SelectedMailbox,
                            last_mod_seq: bytes) -> None:
        keys = self._keys
        redis = self._redis
        await redis.xread({keys.changes: last_mod_seq}, block=1000, count=1)


class MailboxSet(MailboxSetInterface[MailboxData]):
    """Implementation of :class:`~pymap.backend.mailbox.MailboxSetInterface`
    for the redis backend.

    """

    def __init__(self, redis: Redis, keys: NamespaceKeys,
                 cl_keys: CleanupKeys) -> None:
        super().__init__()
        self._redis = redis
        self._keys = keys
        self._cl_keys = cl_keys

    @property
    def delimiter(self) -> str:
        return '/'

    async def set_subscribed(self, name: str, subscribed: bool) -> None:
        if subscribed:
            await self._redis.sadd(self._keys.subscribed, name)
        else:
            await self._redis.srem(self._keys.subscribed, name)

    async def list_subscribed(self) -> ListTree:
        mailboxes = await self._redis.smembers(self._keys.subscribed)
        return ListTree(self.delimiter).update('INBOX', *(
            modutf7_decode(name) for name in mailboxes))

    async def list_mailboxes(self) -> ListTree:
        mailboxes = await _ns_scripts.list(self._redis, self._keys)
        return ListTree(self.delimiter).update('INBOX', *(
            modutf7_decode(name) for name in mailboxes))

    async def get_mailbox(self, name: str) -> MailboxData:
        redis = self._redis
        name_key = modutf7_encode(name)
        try:
            mbx_id, uid_val = await _ns_scripts.get(
                redis, self._keys, name=name_key)
        except ResponseError as exc:
            if 'mailbox not found' in str(exc):
                raise KeyError(name_key) from exc
            raise
        mbx_keys = MailboxKeys(self._keys, mbx_id)
        return MailboxData(redis, mbx_id, uid_val, mbx_keys, self._keys,
                           self._cl_keys)

    async def add_mailbox(self, name: str) -> ObjectId:
        name_key = modutf7_encode(name)
        mbx_id = ObjectId.random_mailbox_id()
        try:
            await _ns_scripts.add(self._redis, self._keys,
                                  name=name_key,
                                  mailbox_id=mbx_id.value)
        except ResponseError as exc:
            if 'mailbox already exists' in str(exc):
                raise ValueError(name_key) from exc
            raise
        return mbx_id

    async def delete_mailbox(self, name: str) -> None:
        name_key = modutf7_encode(name)
        try:
            await _ns_scripts.delete(self._redis, self._keys, self._cl_keys,
                                     name=name_key)
        except ResponseError as exc:
            if 'mailbox not found' in str(exc):
                raise KeyError(name_key) from exc
            raise

    async def rename_mailbox(self, before: str, after: str) -> None:
        redis = self._redis
        async with redis.pipeline() as pipe:
            await pipe.watch(self._keys.mailboxes)
            pipe.multi()
            pipe.hgetall(self._keys.mailboxes)
            all_keys, = await pipe.execute()
        all_mbx = {modutf7_decode(key): ns for key, ns in all_keys.items()}
        tree = ListTree(self.delimiter).update('INBOX', *all_mbx.keys())
        before_entry = tree.get(before)
        after_entry = tree.get(after)
        if before_entry is None:
            raise MailboxNotFound(before)
        elif after_entry is not None:
            raise MailboxConflict(after)
        async with redis.pipeline() as multi:
            for before_name, after_name in tree.get_renames(before, after):
                before_id = all_mbx[before_name]
                before_key = modutf7_encode(before_name)
                after_key = modutf7_encode(after_name)
                multi.hset(self._keys.mailboxes, after_key, before_id)
                multi.hdel(self._keys.mailboxes, before_key)
                multi.hincrby(self._keys.uid_validity, after_key)
            if before == 'INBOX':
                inbox_id = ObjectId.random_mailbox_id()
                multi.hset(self._keys.mailboxes, b'INBOX', inbox_id.value)
                multi.hincrby(self._keys.uid_validity, b'INBOX')
            try:
                await multi.execute()
            except WatchError as exc:
                msg = 'Mailboxes changed externally, please try again.'
                raise TemporaryFailure(msg) from exc




from __future__ import annotations

import enum
from abc import ABCMeta
from collections.abc import Iterable, Sequence
from functools import total_ordering, reduce
from typing import Optional, Final

from . import AString
from .. import Params, Parseable
from ..exceptions import NotParseable
from ..primitives import Atom, List
from ...bytes import rev, BytesFormat, MaybeBytes, Writeable

__all__ = ['FetchRequirement', 'FetchAttribute', 'FetchValue']


class FetchRequirement(enum.Flag):
    """Indicates the data required to fulfill a message fetch.

    Attributes:
        NONE: No data is required.
        METADATA: The IMAP metadata is required.
        HEADER: The message header is required.
        BODY: The parsed MIME message body is required.

    """

    NONE = 0
    METADATA = enum.auto()
    HEADER = enum.auto()
    BODY = enum.auto()
    CONTENT = HEADER | BODY

    def has_all(self, expected: FetchRequirement) -> bool:
        """Returns true if all of the expected fetch requirements are met by
        this fetch requirement.

        Args:
            expected: The expected fetch requirements.

        """
        return self & expected == expected

    def has_none(self, expected: FetchRequirement) -> bool:
        """Returns true if none of the expected fetch requirements are met by
        this fetch requirement.

        Args:
            expected: The expected fetch requirements.

        """
        return self & expected == FetchRequirement.NONE

    @classmethod
    def get_all(cls) -> FetchRequirement:
        """Return all possible fetch requirements reduced into a single
        requirement.

        """
        return cls.reduce(FetchRequirement)

    @classmethod
    def reduce(cls, requirements: Iterable[FetchRequirement]) \
            -> FetchRequirement:
        """Reduce a set of fetch requirements into a single requirement.

        Args:
            requirements: The set of fetch requirements.

        """
        return reduce(lambda x, y: x | y, requirements, cls.NONE)


@total_ordering
class FetchAttribute(Parseable[bytes]):
    """Represents an attribute that should be fetched for each message in the
    sequence set of a FETCH command on an IMAP stream.

    Args:
        attribute: The attribute name.
        section: The attribute section.
        partial: The attribute partial range.

    Attributes:
        section: The attribute section.
        partial: The attribute partial range.

    """

    class Section:
        """Represents a fetch attribute section, which typically comes after
        the attribute name within square brackets, e.g. ``BODY[1.TEXT]``.

        Args:
            parts: The nested MIME part identifiers.
            specifier: The MIME part specifier.
            headers: The MIME part specifier headers.

        """

        def __init__(self, parts: Sequence[int],
                     specifier: bytes = None,
                     headers: frozenset[bytes] = None) -> None:
            self.parts = parts
            self.specifier = specifier
            self.headers = frozenset(hdr.upper() for hdr in headers) \
                if headers else None

        def __hash__(self) -> int:
            return hash((tuple(self.parts), self.specifier, self.headers))

    _attrname_pattern = rev.compile(br' *([^\s\[<()]+)')
    _section_start_pattern = rev.compile(br' *\[ *')
    _section_end_pattern = rev.compile(br' *\]')
    _partial_pattern = rev.compile(br'< *(\d+) *\. *(\d+) *>')

    _sec_part_pattern = rev.compile(br'([1-9]\d* *(?:\. *[1-9]\d*)*) *(\.)? *')

    def __init__(self, attribute: bytes,
                 section: FetchAttribute.Section = None,
                 partial: tuple[int, Optional[int]] = None) -> None:
        super().__init__()
        self.attribute = attribute.upper()
        self.section = section
        self.partial = partial
        self._raw: Optional[bytes] = None
        self._for_response: Optional[FetchAttribute] = None

    @property
    def value(self) -> bytes:
        """The attribute name."""
        return self.attribute

    @property
    def for_response(self) -> FetchAttribute:
        if self._for_response is None:
            if self.partial is None or len(self.partial) < 2:
                self._for_response = self
            else:
                self._for_response = FetchAttribute(
                    self.value, self.section, (self.partial[0], None))
        return self._for_response

    @property
    def set_seen(self) -> bool:
        if self.value == b'BODY' and self.section:
            return True
        elif self.value == b'BINARY':
            return True
        elif self.value in (b'RFC822', b'RFC822.TEXT'):
            return True
        return False

    @property
    def requirement(self) -> FetchRequirement:
        """Indicates the data required to fulfill this fetch attribute."""
        attr_name = self.attribute
        if attr_name in (b'UID', b'FLAGS', b'INTERNALDATE'):
            return FetchRequirement.METADATA
        elif attr_name in (b'ENVELOPE', b'RFC822.HEADER'):
            return FetchRequirement.HEADER
        elif attr_name == b'BODY' and self.section is not None:
            return self._get_body_requirement(self.section)
        return FetchRequirement.CONTENT

    @classmethod
    def _get_body_requirement(cls, section: FetchAttribute.Section) \
            -> FetchRequirement:
        if not section.parts and section.specifier in (
                b'HEADER', b'HEADER.FIELDS', b'HEADER.FIELDS.NOT'):
            return FetchRequirement.HEADER
        else:
            return FetchRequirement.CONTENT

    @property
    def raw(self) -> bytes:
        if self._raw is not None:
            return self._raw
        if self.value == b'BODY.PEEK':
            parts = [b'BODY']
        else:
            parts = [self.value]
        if self.section and not self.value.startswith(b'RFC822'):
            parts.append(b'[')
            if self.section.parts:
                part_raw = BytesFormat(b'.').join(
                    [b'%i' % num for num in self.section.parts])
                parts.append(part_raw)
                if self.section.specifier:
                    parts.append(b'.')
            if self.section.specifier:
                parts.append(self.section.specifier)
                if self.section.headers:
                    headers = self.section.headers
                    parts.append(b' ')
                    parts.append(bytes(List(headers, sort=True)))
            parts.append(b']')
        if self.partial:
            partial = BytesFormat(b'.').join(
                [b'%i' % num for num in self.partial if num is not None])
            parts += [b'<', partial, b'>']
        self._raw = raw = b''.join(parts)
        return raw

    def __hash__(self) -> int:
        return hash((self.value, self.section, self.partial))

    def __eq__(self, other) -> bool:
        if isinstance(other, FetchAttribute):
            return hash(self) == hash(other)
        return super().__eq__(other)

    def __ne__(self, other) -> bool:
        if isinstance(other, FetchAttribute):
            return hash(self) != hash(other)
        return super().__ne__(other)

    def __lt__(self, other) -> bool:
        if not isinstance(other, FetchAttribute):
            return NotImplemented
        return bytes(self.for_response) < bytes(self.for_response)

    @classmethod
    def _parse_section(cls, buf: memoryview, params: Params):
        match = cls._sec_part_pattern.match(buf)
        if match:
            section_parts = [int(num) for num in match.group(1).split(b'.')]
            buf = buf[match.end(0):]
        else:
            section_parts = []
        try:
            atom, after = Atom.parse(buf, params)
        except NotParseable:
            return cls.Section(section_parts), buf
        specifier = atom.value.upper()
        if section_parts and specifier == b'MIME':
            return cls.Section(section_parts, specifier), after
        elif specifier in (b'HEADER', b'TEXT'):
            return cls.Section(section_parts, specifier), after
        elif specifier in (b'HEADER.FIELDS', b'HEADER.FIELDS.NOT'):
            params = params.copy(list_expected=[AString])
            header_list_p, buf = List.parse(after, params)
            header_list = frozenset([bytes(hdr)
                                     for hdr in header_list_p.value])
            if not header_list:
                raise NotParseable(after)
            return cls.Section(section_parts, specifier, header_list), buf
        raise NotParseable(buf)

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[FetchAttribute, memoryview]:
        match = cls._attrname_pattern.match(buf)
        if not match:
            raise NotParseable(buf)
        attr = match.group(1).upper()
        after = buf[match.end(0):]
        if attr in (b'ENVELOPE', b'FLAGS', b'INTERNALDATE', b'UID',
                    b'RFC822.SIZE',  b'BODYSTRUCTURE', b'EMAILID',
                    b'THREADID'):
            return cls(attr), after
        elif attr == b'RFC822':
            section = cls.Section([])
            return cls(attr, section), after
        elif attr == b'RFC822.HEADER':
            section = cls.Section([], b'HEADER')
            return cls(attr, section), after
        elif attr == b'RFC822.TEXT':
            section = cls.Section([], b'TEXT')
            return cls(attr, section), after
        elif attr not in (b'BODY', b'BODY.PEEK',
                          b'BINARY', b'BINARY.PEEK', b'BINARY.SIZE'):
            raise NotParseable(buf)
        buf = after
        match = cls._section_start_pattern.match(buf)
        if not match:
            if attr == b'BODY':
                return cls(attr), buf
            else:
                raise NotParseable(buf)
        buf = buf[match.end(0):]
        section, buf_s = cls._parse_section(buf, params)
        if section.specifier and attr.startswith(b'BINARY'):
            raise NotParseable(buf)
        match = cls._section_end_pattern.match(buf_s)
        if not match:
            raise NotParseable(buf_s)
        buf = buf_s[match.end(0):]
        match = cls._partial_pattern.match(buf)
        if match:
            if attr == b'BINARY.SIZE':
                raise NotParseable(buf)
            from_, to = int(match.group(1)), int(match.group(2))
            if from_ < 0 or to <= 0 or from_ > to:
                raise NotParseable(buf)
            return cls(attr, section, (from_, to)), buf[match.end(0):]
        return cls(attr, section), buf

    def __bytes__(self) -> bytes:
        return self.raw


class FetchValue(Writeable, metaclass=ABCMeta):
    """A value fetched from a message for a single fetch attribute.

    Args:
        attribute: The fetch attribute.

    """

    __slots__ = ['attribute']

    def __init__(self, attribute: FetchAttribute) -> None:
        super().__init__()
        self.attribute: Final = attribute

    @classmethod
    def of(cls, attribute: FetchAttribute, value: MaybeBytes) \
            -> FetchValue:
        return _StaticFetchValue(attribute, value)


class _StaticFetchValue(FetchValue):

    __slots__ = ['_value']

    def __init__(self, attribute: FetchAttribute, value: MaybeBytes) -> None:
        super().__init__(attribute)
        self._value: Final = value

    def __bytes__(self) -> bytes:
        attr = self.attribute.for_response
        return BytesFormat(b'%b %b') % (attr, self._value)




from __future__ import annotations

import re
from collections.abc import Set
from datetime import datetime, timezone
from typing import Any, Final, Optional, NoReturn

from pymacaroons import Macaroon, Verifier
from pymacaroons.exceptions import MacaroonDeserializationException, \
    MacaroonInvalidSignatureException
from pysasl.creds import AuthenticationCredentials, StoredSecret

from ..interfaces.token import TokensInterface

__all__ = ['MacaroonTokens', 'MacaroonCredentials']


class MacaroonTokens(TokensInterface):
    """Creates and parses :class:`~pymacaroons.Macaroon` tokens."""

    def get_login_token(self, identifier: str, key: bytes, *,
                        authzid: str = None, location: str = None,
                        expiration: datetime = None) -> str:
        macaroon = Macaroon(location=location, identifier=identifier, key=key)
        macaroon.add_first_party_caveat('type = login')
        if authzid is not None:
            macaroon.add_first_party_caveat(f'authzid = {authzid}')
        if expiration is not None:
            macaroon.add_first_party_caveat(f'time < {expiration.isoformat()}')
        return macaroon.serialize()

    def get_admin_token(self, admin_key: Optional[bytes], *,
                        authzid: str = None, location: str = None,
                        expiration: datetime = None) -> Optional[str]:
        if admin_key is None:
            return None
        macaroon = Macaroon(location=location, identifier='', key=admin_key)
        macaroon.add_first_party_caveat('type = admin')
        if authzid is not None:
            macaroon.add_first_party_caveat(f'authzid = {authzid}')
        if expiration is not None:
            macaroon.add_first_party_caveat(f'time < {expiration.isoformat()}')
        return macaroon.serialize()

    def parse(self, authzid: str, token: str, *,
              admin_keys: Set[bytes] = frozenset()) -> MacaroonCredentials:
        try:
            return MacaroonCredentials(authzid, token, admin_keys)
        except MacaroonDeserializationException as exc:
            raise ValueError('invalid macaroon') from exc


class MacaroonCredentials(AuthenticationCredentials):
    """Authenticate using a `Macaroon`_ token. Tokens may be created with
    either :meth:`~MacaroonTokens.get_login_token` or
    :meth:`~MacaroonTokens.get_admin_token`.

    Either type of token may use a caveat with ``time <`` and a
    :func:`~datetime.datetime.fromisoformat` string to limit how long the token
    is valid.

    .. _Macaroon: https://github.com/ecordell/pymacaroons#readme

    Args:
        identity: The identity to be authorized.
        serialized: The serialized macaroon string.
        admin_keys: The admin token string.

    """

    _caveat = re.compile(r'^(\w+) ([^\w\s]+) (.*)$', re.ASCII)

    def __init__(self, identity: str, serialized: str,
                 admin_keys: Set[bytes]) -> None:
        macaroon = Macaroon.deserialize(serialized)
        authcid_type = self._find_type(macaroon)
        super().__init__(macaroon.identifier, '', identity,
                         authcid_type=authcid_type)
        self.macaroon: Final = macaroon
        self.admin_keys: Final = admin_keys

    @classmethod
    def _find_type(cls, macaroon: Macaroon) -> Optional[str]:
        for caveat in macaroon.first_party_caveats():
            caveat_id = caveat.caveat_id
            if caveat_id == 'type = admin':
                return 'admin-token'
            elif caveat_id == 'type = login':
                return 'login-token'
        return None

    @property
    def has_secret(self) -> bool:
        return False

    @property
    def secret(self) -> NoReturn:
        raise AttributeError('secret')

    def _satisfy(self, predicate: str) -> bool:
        match = self._caveat.match(predicate)
        if match is None:
            return False
        key, op, value = match.groups()
        if key == 'time' and op == '<':
            try:
                end = datetime.fromisoformat(value)
            except ValueError:
                return False
            now = datetime.now(tz=timezone.utc)
            return now < end
        elif key == 'authzid' and op == '=':
            return value == self.authzid
        else:
            return False

    def _verify(self, verifier: Verifier, key: bytes) -> bool:
        try:
            return verifier.verify(self.macaroon, key)
        except MacaroonInvalidSignatureException:
            return False

    def _get_login_verifier(self) -> Verifier:
        verifier = Verifier()
        verifier.satisfy_general(self._satisfy)
        verifier.satisfy_exact('type = login')
        return verifier

    def _get_admin_verifier(self) -> Verifier:
        verifier = Verifier()
        verifier.satisfy_general(self._satisfy)
        verifier.satisfy_exact('type = admin')
        return verifier

    def check_secret(self, secret: Optional[StoredSecret], *,
                     key: bytes = None, **other: Any) -> bool:
        if key is not None:
            verifier = self._get_login_verifier()
            if self._verify(verifier, key):
                return True
        for admin_key in self.admin_keys:
            verifier = self._get_admin_verifier()
            if self._verify(verifier, admin_key):
                return True
        return False




from __future__ import annotations

import os
import os.path
from abc import abstractmethod, ABCMeta
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager, AbstractAsyncContextManager
from tempfile import NamedTemporaryFile
from typing import TypeVar, Generic, Optional, IO

from pymap.concurrent import FileLock

__all__ = ['NoChanges', 'FileReadable', 'FileWriteable']

_RT = TypeVar('_RT', bound='FileReadable')
_WT = TypeVar('_WT', bound='FileWriteable')


class NoChanges(Exception):
    pass


class FileReadable(metaclass=ABCMeta):

    @classmethod
    @abstractmethod
    def get_file(cls) -> str:
        ...

    @classmethod
    def get_lock(cls) -> Optional[str]:
        return None

    @classmethod
    @abstractmethod
    def get_default(cls: type[_RT], base_dir: str) -> _RT:
        ...

    @classmethod
    @abstractmethod
    def open(cls: type[_RT], base_dir: str, fp: IO[str]) -> _RT:
        ...

    def read(self, fp: IO[str]) -> None:
        pass

    @classmethod
    @asynccontextmanager
    async def _noop_lock(cls) -> AsyncIterator[None]:
        yield

    @classmethod
    def read_lock(cls, base_dir: str) -> AbstractAsyncContextManager[None]:
        lock_file = cls.get_lock()
        if lock_file:
            path = os.path.join(base_dir, lock_file)
            return FileLock(path).read_lock()
        else:
            return cls._noop_lock()

    @classmethod
    def write_lock(cls, base_dir: str) -> AbstractAsyncContextManager[None]:
        lock_file = cls.get_lock()
        if lock_file:
            path = os.path.join(base_dir, lock_file)
            return FileLock(path).write_lock()
        else:
            return cls._noop_lock()

    @classmethod
    def file_exists(cls, base_dir: str) -> bool:
        path = os.path.join(base_dir, cls.get_file())
        return os.path.exists(path)

    @classmethod
    def file_open(cls: type[_RT], base_dir: str) -> _RT:
        path = os.path.join(base_dir, cls.get_file())
        try:
            with open(path, 'r') as in_file:
                return cls.open(base_dir, in_file)
        except FileNotFoundError:
            return cls.get_default(base_dir)

    @classmethod
    def file_read(cls: type[_RT], base_dir: str) -> _RT:
        path = os.path.join(base_dir, cls.get_file())
        try:
            with open(path, 'r') as in_file:
                ret = cls.open(base_dir, in_file)
                ret.read(in_file)
                return ret
        except FileNotFoundError:
            return cls.get_default(base_dir)

    @classmethod
    def with_open(cls: type[_RT], base_dir: str) -> _FileReadWith[_RT]:
        return _FileReadWith(base_dir, cls, True)

    @classmethod
    def with_read(cls: type[_RT], base_dir: str) -> _FileReadWith[_RT]:
        return _FileReadWith(base_dir, cls, False)


class FileWriteable(FileReadable, metaclass=ABCMeta):

    @abstractmethod
    def get_dir(self) -> str:
        ...

    @abstractmethod
    def write(self, fp: IO[str]) -> None:
        ...

    @classmethod
    def delete(cls, base_dir: str) -> None:
        path = os.path.join(base_dir, cls.get_file())
        try:
            os.unlink(path)
        except OSError:
            pass

    def file_write(self) -> None:
        filename = self.get_file()
        base_dir = self.get_dir()
        tmp_dir = os.path.join(base_dir, 'tmp')
        path = os.path.join(base_dir, filename)
        with NamedTemporaryFile('w', dir=tmp_dir, delete=False) as tmp:
            self.write(tmp)
        tmp_path = os.path.join(base_dir, tmp.name)
        os.rename(tmp_path, path)

    @classmethod
    def with_write(cls: type[_WT], base_dir: str) -> _FileWriteWith[_WT]:
        return _FileWriteWith(base_dir, cls)


class _FileReadWith(Generic[_RT]):

    def __init__(self, base_dir: str, cls: type[_RT], only_open: bool) -> None:
        super().__init__()
        self._base_dir = base_dir
        self._cls = cls
        self._obj: Optional[_RT] = None
        self._only_open = only_open

    async def __aenter__(self) -> _RT:
        base_dir = self._base_dir
        cls = self._cls
        func = cls.file_open if self._only_open else cls.file_read
        async with cls.read_lock(base_dir):
            self._obj = obj = func(base_dir)
        return obj

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> bool:
        return False


class _FileWriteWith(Generic[_WT]):

    def __init__(self, base_dir: str, cls: type[_WT]) -> None:
        super().__init__()
        self._base_dir = base_dir
        self._cls = cls
        self._exists = False
        self._lock: Optional[AbstractAsyncContextManager[None]] = None
        self._obj: Optional[_WT] = None

    async def _acquire_lock(self) -> None:
        self._lock = self._cls.write_lock(self._base_dir)
        await self._lock.__aenter__()

    async def _release_lock(self) -> None:
        lock = self._lock
        if not lock:
            raise RuntimeError()  # Lock not acquired.
        await lock.__aexit__(None, None, None)

    async def __aenter__(self) -> _WT:
        base_dir = self._base_dir
        cls = self._cls
        await self._acquire_lock()
        self._exists = cls.file_exists(base_dir)
        self._obj = obj = cls.file_read(base_dir)
        return obj

    def _write_obj(self) -> None:
        obj = self._obj
        if not obj:
            raise RuntimeError()  # Context manager not entered.
        obj.file_write()

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> bool:
        try:
            if exc_type:
                if issubclass(exc_type, NoChanges):
                    if not self._exists:
                        self._write_obj()
                    return True
                else:
                    return False
            self._write_obj()
            return False
        finally:
            await self._release_lock()




from __future__ import annotations

from abc import abstractmethod
from collections.abc import Mapping
from typing import Protocol

from grpclib.const import Handler as _Handler

__all__ = ['Handler']


class Handler(Protocol):
    """Defines the protocol for :class:`~grpclib.server.Server` handlers. This
    can be removed if ``grpclib._typing.IServable`` becomes public API.

    """

    @abstractmethod
    def __mapping__(self) -> Mapping[str, _Handler]:
        ...




from __future__ import annotations

from collections.abc import Mapping
from typing import Optional, Final

from pysasl.creds import StoredSecret, AuthenticationCredentials

from .config import IMAPConfig
from .exceptions import InvalidAuth

__all__ = ['UserMetadata']


class UserMetadata:
    """Contains user metadata such as the password or hash.

    Args:
        config: The configuration object.
        password: The password string or hash digest.
        params: Metadata parameters associated with the user.

    """

    def __init__(self, config: IMAPConfig, *, password: str = None,
                 **params: Optional[str]) -> None:
        super().__init__()
        self.config: Final = config
        self.password: Final = password
        self.params: Final = {key: val for key, val in params.items()
                              if val is not None}

    @property
    def role(self) -> Optional[str]:
        """The value of the ``role`` key from *params*."""
        return self.params.get('role')

    def to_dict(self, **extra: str) -> Mapping[str, str]:
        """Combines the *password*, *params*, and *extra* into a dictionary.

        Args:
            extra: Additional parameters as keyword arguments, which will be
                merged into the result.

        """
        ret = dict(self.params, **extra)
        if self.password is not None:
            ret['password'] = self.password
        return ret

    async def check_password(self, creds: AuthenticationCredentials, *,
                             token_key: bytes = None) -> None:
        """Check the given credentials against the known password comparison
        data. If the known data used a hash, then the equivalent hash of the
        provided secret is compared.

        Args:
            creds: The provided authentication credentials.
            token_key: The token key bytestring.

        Raises:
            :class:`~pymap.exceptions.InvalidAuth`

        """
        hash_context = self.config.hash_context.copy()
        cpu_subsystem = self.config.cpu_subsystem
        stored_secret: Optional[StoredSecret] = None
        if self.password is not None:
            stored_secret = StoredSecret(self.password, hash=hash_context)
        fut = self._check_secret(creds, stored_secret, token_key)
        if not await cpu_subsystem.execute(fut):
            raise InvalidAuth()

    async def _check_secret(self, creds: AuthenticationCredentials,
                            stored_secret: Optional[StoredSecret],
                            key: Optional[bytes]) -> bool:
        return creds.check_secret(stored_secret, key=key)




from __future__ import annotations

from collections.abc import Set
from datetime import datetime
from typing import Optional

from pysasl.creds import AuthenticationCredentials

from ..interfaces.token import TokensInterface
from ..plugin import Plugin

__all__ = ['tokens', 'AllTokens']

#: Registers token plugins.
tokens: Plugin[type[TokensInterface]] = Plugin(
    'pymap.token', default='macaroon')


class AllTokens(TokensInterface):
    """Uses :data:`tokens` to support all registered token types.

    For token creation, the :attr:`~pymap.plugin.Plugin.default` token plugin
    is used. For token parsing, each token plugin is tried until one succeeds.

    """

    def get_login_token(self, identifier: str, key: bytes, *,
                        authzid: str = None, location: str = None,
                        expiration: datetime = None) -> Optional[str]:
        try:
            token_type = tokens.default
        except KeyError:
            return None
        return token_type().get_login_token(
            identifier, key, authzid=authzid, location=location,
            expiration=expiration)

    def get_admin_token(self, admin_key: Optional[bytes], *,
                        authzid: str = None, location: str = None,
                        expiration: datetime = None) -> Optional[str]:
        try:
            token_type = tokens.default
        except KeyError:
            return None
        return token_type().get_admin_token(
            admin_key, authzid=authzid, location=location,
            expiration=expiration)

    def parse(self, authzid: str, token: str, *,
              admin_keys: Set[bytes] = frozenset()) \
            -> AuthenticationCredentials:
        for _, token_type in tokens.registered.items():
            try:
                return token_type().parse(authzid, token,
                                          admin_keys=admin_keys)
            except ValueError:
                pass
        raise ValueError('invalid token')




from __future__ import annotations

from collections.abc import Iterable
from datetime import datetime
from typing import Optional

import msgpack
from aioredis import Redis

from pymap.message import BaseMessage, BaseLoadedMessage
from pymap.mime import MessageContent, MessageHeader, MessageBody
from pymap.parsing.specials import Flag, ObjectId, FetchRequirement

from .keys import NamespaceKeys, ContentKeys

__all__ = ['Message', 'LoadedMessage']


class Message(BaseMessage):

    __slots__ = ['_redis', '_ns_keys']

    def __init__(self, uid: int, internal_date: datetime,
                 permanent_flags: Iterable[Flag], *, expunged: bool = False,
                 email_id: ObjectId = None, thread_id: ObjectId = None,
                 redis: Redis = None, ns_keys: NamespaceKeys = None) -> None:
        super().__init__(uid, internal_date, permanent_flags,
                         expunged=expunged, email_id=email_id,
                         thread_id=thread_id)
        self._redis = redis
        self._ns_keys = ns_keys

    async def _load_full(self, redis: Redis, ct_keys: ContentKeys) \
            -> MessageContent:
        literal, full_json_raw = await redis.hmget(
            ct_keys.data, b'full', b'full-json')
        if literal is None or full_json_raw is None:
            raise ValueError(f'Missing message content: {self.email_id}')
        full_json = msgpack.unpackb(full_json_raw, raw=False)
        return MessageContent.from_json(literal, full_json)

    async def _load_header(self, redis: Redis, ct_keys: ContentKeys) \
            -> MessageContent:
        literal, header_json_raw = await redis.hmget(
            ct_keys.data, b'header', b'header-json')
        if literal is None or header_json_raw is None:
            raise ValueError(f'Missing message header: {self.email_id}')
        header_json = msgpack.unpackb(header_json_raw, raw=False)
        header = MessageHeader.from_json(literal, header_json)
        body = MessageBody.empty()
        return MessageContent(literal, header, body)

    async def load_content(self, requirement: FetchRequirement) \
            -> LoadedMessage:
        redis = self._redis
        ns_keys = self._ns_keys
        if redis is None or ns_keys is None:
            return LoadedMessage(self, requirement, None)
        ct_keys = ContentKeys(ns_keys, self.email_id)
        content: Optional[MessageContent] = None
        if requirement & FetchRequirement.BODY:
            content = await self._load_full(redis, ct_keys)
        elif requirement & FetchRequirement.HEADER:
            content = await self._load_header(redis, ct_keys)
        return LoadedMessage(self, requirement, content)

    @classmethod
    def copy_expunged(cls, msg: Message) -> Message:
        return cls(msg.uid, msg.internal_date, msg.permanent_flags,
                   expunged=True, email_id=msg.email_id,
                   thread_id=msg.thread_id, redis=msg._redis,
                   ns_keys=msg._ns_keys)


class LoadedMessage(BaseLoadedMessage):
    pass




from __future__ import annotations

import logging
from collections.abc import Callable
from types import MethodType
from typing import Any, Optional
from weakref import finalize, ref, WeakMethod, WeakSet

__all__ = ['HealthStatus']

_log = logging.getLogger(__name__)

_Callback = Callable[[bool], Any]
_WeakCallback = Callable[[], Optional[_Callback]]


class HealthStatus:
    """Provides a mechanism for backends or services to track their health as a
    simple boolean state: healthy or unhealthy.

    Args:
        initial: The initial health status.
        name: A name for the health status component.

    """

    def __init__(self, initial: bool = True, *, name: str = '') -> None:
        super().__init__()
        self._name = name
        self._listeners: set[_WeakCallback] = set()
        self._watchers: WeakSet[HealthStatus] = WeakSet()
        self._dependencies: WeakSet[HealthStatus] = WeakSet()
        self._healthy = initial
        self._prev_healthy = initial
        if name:
            self._listeners.add(lambda: self._log_change)

    @property
    def name(self) -> str:
        """The name of the health status component."""
        return self._name

    @property
    def healthy(self) -> bool:
        """The current healthy (True) or unhealthy (False) status."""
        return self._healthy and all(dep.healthy for dep in self._dependencies)

    def _log_change(self, healthy: bool) -> None:
        health_str = 'healthy' if healthy else 'unhealthy'
        _log.debug('%s is now %s', self.name, health_str)

    def register(self, callback: _Callback) -> None:
        """Registers a callback to be called when transitioning from healthy to
        unhealthy or from unhealthy to healthy. The callback will also be
        called immediately with the current value of :attr:`.healthy`.

        Args:
            callback: Takes a single boolean argument, :attr:`.healthy`.

        """
        if isinstance(callback, MethodType):
            self._listeners.add(WeakMethod(callback))
        else:
            self._listeners.add(ref(callback))
        callback(self.healthy)

    def _check(self, seen: set[HealthStatus]) -> None:
        new_healthy = self.healthy
        if new_healthy != self._prev_healthy:
            self._prev_healthy = new_healthy
            seen.add(self)
            for listener in self._listeners:
                cb = listener()
                if cb is not None:
                    cb(new_healthy)
            for watcher in self._watchers:
                if watcher not in seen:
                    watcher._check(seen)

    def new_dependency(self, initial: bool = True, *,
                       name: str = '') -> HealthStatus:
        """Adds a new health status as a dependency of this status. All
        dependencies must be :attr:`.healthy` for this status to be
        :attr:`.healthy`.

        Args:
            initial: The initial health status of the dependency.
            name: A name for the health status component.

        """
        full_name = f'{self.name}.{name}' if name else ''
        status = HealthStatus(initial, name=full_name)
        self._dependencies.add(status)
        status._watchers.add(self)
        self._check(set())
        finalize(status, self._check, set())
        return status

    def set(self, healthy: bool) -> None:
        """Updates healthy or unhealthy.

        Args:
            healthy: The new healthy (True) or unhealthy (False) status.

        """
        self._healthy = healthy
        self._check(set())

    def set_healthy(self) -> None:
        """Shortcut for calling :meth:`.set` with True."""
        self.set(True)

    def set_unhealthy(self) -> None:
        """Shortcut for calling :meth:`.set` with False."""
        self.set(False)




from __future__ import annotations

import asyncio
import binascii
import logging
import re
from argparse import ArgumentParser
from asyncio import StreamReader, StreamWriter
from base64 import b64encode, b64decode
from collections.abc import Mapping
from contextlib import closing, AsyncExitStack
from typing import Optional, Union

from proxyprotocol import ProxyProtocolResult
from proxyprotocol.reader import ProxyProtocolReader
from proxyprotocol.sock import SocketInfo
from pymap import __version__
from pymap.bytes import BytesFormat
from pymap.config import IMAPConfig
from pymap.context import socket_info, language_code, connection_exit
from pymap.exceptions import InvalidAuth
from pymap.interfaces.backend import ServiceInterface
from pymap.interfaces.login import LoginInterface
from pymap.interfaces.session import SessionInterface
from pymap.parsing.exceptions import NotParseable
from pymap.parsing.primitives import String
from pysasl import ServerChallenge, ChallengeResponse, AuthenticationError, \
    AuthenticationCredentials

from .command import Command, NoOpCommand, LogoutCommand, CapabilityCommand, \
    AuthenticateCommand, UnauthenticateCommand, StartTLSCommand
from .response import Condition, Response, BadCommandResponse, NoOpResponse, \
    CapabilitiesResponse
from .state import FilterState
from .. import SieveCompiler

__all__ = ['ManageSieveService', 'ManageSieveServer', 'ManageSieveConnection']

_log = logging.getLogger(__name__)


class ManageSieveService(ServiceInterface):  # pragma: no cover
    """A pymap service that implements a ManageSieve server to control the
    sieve scripts associated with a login.

    See Also:
        `RFC 5804 <https://tools.ietf.org/html/rfc5804>`_,
        :class:`~pymap.interfaces.filter.FilterSetInterface`

    """

    @classmethod
    def add_arguments(cls, parser: ArgumentParser) -> None:
        group = parser.add_argument_group('managesieve service')
        group.add_argument('--sieve-host', action='append', metavar='IFACE',
                           help='the network interface to listen on')
        group.add_argument('--sieve-port',  metavar='NUM', default='4190',
                           help='the port or service name to listen on')

    async def start(self, stack: AsyncExitStack) -> None:
        backend = self.backend
        config = self.config
        managesieve_server = ManageSieveServer(backend.login, config)
        host: Optional[str] = config.args.sieve_host
        port: Union[str, int] = config.args.sieve_port
        server = await asyncio.start_server(
            managesieve_server, host=host, port=port)
        await stack.enter_async_context(server)
        task = asyncio.create_task(server.serve_forever())
        stack.callback(task.cancel)


class ManageSieveServer:
    """Callable object that creates and runs :class:`ManageSieveConnection`
    objects when :func:`asyncio.start_server` receives a new connection.

    Args:
        login: Login callback that takes authentication credentials and returns
            a :class:`~pymap.interfaces.session.SessionInterface` object.
        config: Settings to use for the sieve server.

    """

    def __init__(self, login: LoginInterface, config: IMAPConfig) -> None:
        super().__init__()
        self._login = login
        self._config = config

    async def __call__(self, reader: StreamReader,
                       writer: StreamWriter) -> None:
        conn = ManageSieveConnection(self._login, self._config, reader, writer)
        async with AsyncExitStack() as stack:
            connection_exit.set(stack)
            stack.enter_context(closing(writer))
            await conn.run()


class ManageSieveConnection:
    """Runs a single ManageSieve connection from start to finish.

    Args:
        login: Login callback that takes authentication credentials and returns
            a :class:`~pymap.interfaces.session.SessionInterface` object.
        config: Settings to use for the sieve server.
        reader: The input stream for the socket.
        writer: The output stream for the socket.

    """

    _lines = re.compile(r'\r?\n')
    _literal_plus = re.compile(br'{(\d+)\+}\r?\n$')
    _impl = b'pymap managesieve ' + __version__.encode('ascii')

    def __init__(self, login: LoginInterface, config: IMAPConfig,
                 reader: StreamReader, writer: StreamWriter) -> None:
        super().__init__()
        self.login = login
        self.config = config
        self.auth = config.initial_auth
        self.params = config.parsing_params.copy(allow_continuations=False)
        self.pp_reader = ProxyProtocolReader(config.proxy_protocol)
        self.pp_result: Optional[ProxyProtocolResult] = None
        self._offer_starttls = b'STARTTLS' in config.initial_capability
        self._state: Optional[FilterState] = None
        self._reset_streams(reader, writer)

    def _reset_streams(self, reader: StreamReader,
                       writer: StreamWriter) -> None:
        self.reader = reader
        self.writer = writer
        socket_info.set(SocketInfo(writer, self.pp_result))

    async def _read_proxy_protocol(self) -> None:
        self.pp_result = await self.pp_reader.read(self.reader)
        self._reset_streams(self.reader, self.writer)

    def _get_state(self, session: SessionInterface) -> FilterState:
        owner = session.owner.encode('utf-8')
        if session.filter_set is None:
            raise ValueError('Filters not supported.')
        return FilterState(session.filter_set, owner, self.config)

    @property
    def capabilities(self) -> Mapping[bytes, Optional[bytes]]:
        ret: dict[bytes, Optional[bytes]] = {}
        ret[b'IMPLEMENTATION'] = self._impl
        if self._state is None:
            ret[b'SASL'] = b' '.join(
                mech.name for mech in self.auth.server_mechanisms)
        ret[b'SIEVE'] = b' '.join(SieveCompiler.extensions)
        if self._offer_starttls and self._state is None:
            ret[b'STARTTLS'] = None
        try:
            ret[b'LANGUAGE'] = language_code.get().encode('ascii')
        except LookupError:
            pass
        if self._state is not None:
            ret[b'OWNER'] = self._state.owner
        ret[b'UNAUTHENTICATE'] = None
        ret[b'VERSION'] = b'1.0'
        return ret

    @classmethod
    def _print(cls, log_format: str, output: Union[str, bytes]) -> None:
        if _log.isEnabledFor(logging.DEBUG):
            fd = socket_info.get().socket.fileno()
            if not isinstance(output, str):
                output = str(output, 'utf-8', 'replace')
            lines = cls._lines.split(output)
            if not lines[-1]:
                lines = lines[:-1]
            for line in lines:
                _log.debug(log_format, fd, line)

    async def _read_data(self) -> memoryview:
        data = bytearray()
        while True:
            data += await self.reader.readline()
            if not data.endswith(b'\n'):
                raise EOFError()
            match = self._literal_plus.search(data)
            if not match:
                break
            literal_length = int(match.group(1))
            data += await self.reader.readexactly(literal_length)
        self._print('%d -->| %s', data)
        return memoryview(data)

    async def _read_command(self) -> Command:
        cmd_buf = await self._read_data()
        cmd, _ = Command.parse(cmd_buf, self.params)
        return cmd

    async def _write_response(self, resp: Response) -> None:
        resp.write(self.writer)
        try:
            await self.writer.drain()
        except ConnectionError:
            pass
        else:
            self._print('%d <--| %s', bytes(resp))

    async def _login(self, creds: AuthenticationCredentials) \
            -> SessionInterface:
        stack = connection_exit.get()
        identity = await self.login.authenticate(creds)
        return await stack.enter_async_context(identity.new_session())

    async def _do_greeting(self) -> Response:
        sock_info = socket_info.get()
        if self.config.reject_dnsbl and sock_info.dnsbl is not None:
            return Response(Condition.BYE,
                            text=f'Connection rejected: {sock_info.dnsbl}')
        preauth_creds = self.config.preauth_credentials
        if preauth_creds:
            session = await self._login(preauth_creds)
            try:
                self._state = self._get_state(session)
            except ValueError as exc:
                return Response(Condition.NO, text=str(exc))
        return CapabilitiesResponse(self.capabilities)

    async def _do_authenticate(self, cmd: AuthenticateCommand) -> Response:
        mech = self.auth.get_server(cmd.mech_name)
        if not mech:
            return Response(Condition.NO, text='Invalid SASL mechanism.')
        responses: list[ChallengeResponse] = []
        if cmd.initial_data is not None:
            resp_dec = b64decode(cmd.initial_data)
            responses.append(ChallengeResponse(b'', resp_dec))
        while True:
            try:
                creds, final = mech.server_attempt(responses)
            except ServerChallenge as chal:
                chal_bytes = b64encode(chal.data)
                chal_str = String.build(chal_bytes)
                chal_str.write(self.writer)
                self.writer.write(b'\r\n')
                await self.writer.drain()
                resp_bytes = await self._read_data()
                resp_str, _ = String.parse(resp_bytes, self.params)
                if resp_str.value == b'*':
                    raise AuthenticationError('Authentication cancelled.')
                try:
                    resp_dec = b64decode(resp_str.value)
                except binascii.Error as exc:
                    raise AuthenticationError() from exc
                else:
                    responses.append(ChallengeResponse(chal.data, resp_dec))
            except AuthenticationError as exc:
                return Response(Condition.NO, text=str(exc))
            else:
                break
        if final is None:
            code: Optional[bytes] = None
        else:
            code = BytesFormat(b'SASL %b') % String.build(final)
        try:
            session = await self._login(creds)
        except InvalidAuth as exc:
            return Response(Condition.NO, text=str(exc))
        try:
            self._state = self._get_state(session)
        except ValueError as exc:
            return Response(Condition.NO, text=str(exc))
        return Response(Condition.OK, code=code)

    async def _do_unauthenticate(self) -> Response:
        if self._state is None:
            return Response(Condition.NO, text='Not authenticated.')
        else:
            self._state = None
            return Response(Condition.OK)

    async def _do_starttls(self) -> Response:
        ssl_context = self.config.ssl_context
        if not self._offer_starttls:
            return Response(Condition.NO, text='Bad command.')
        resp = Response(Condition.OK)
        await self._write_response(resp)
        loop = asyncio.get_event_loop()
        transport = self.writer.transport
        protocol = transport.get_protocol()
        new_transport = await loop.start_tls(
            transport, protocol, ssl_context, server_side=True)
        new_protocol = new_transport.get_protocol()
        new_writer = StreamWriter(new_transport, new_protocol,
                                  self.reader, loop)
        self._reset_streams(self.reader, new_writer)
        self._print('%d <->| %s', b'<TLS handshake>')
        self._offer_starttls = False
        self.auth = self.config.tls_auth
        return CapabilitiesResponse(self.capabilities)

    async def run(self) -> None:
        """Start the socket communication with the server greeting, and then
        enter the command/response cycle.

        """
        await self._read_proxy_protocol()
        self._print('%d +++| %s', str(socket_info.get()))
        greeting = await self._do_greeting()
        await self._write_response(greeting)
        done = greeting.is_bye
        while not done:
            resp: Response
            try:
                cmd = await self._read_command()
            except (ConnectionError, EOFError):
                break
            except NotParseable as exc:
                resp = BadCommandResponse(exc)
            else:
                try:
                    if isinstance(cmd, NoOpCommand):
                        resp = NoOpResponse(cmd.tag)
                    elif isinstance(cmd, LogoutCommand):
                        resp = Response(Condition.BYE)
                    elif isinstance(cmd, CapabilityCommand):
                        resp = CapabilitiesResponse(self.capabilities)
                    elif self._state is None:
                        if isinstance(cmd, AuthenticateCommand):
                            resp = await self._do_authenticate(cmd)
                        elif isinstance(cmd, StartTLSCommand):
                            resp = await self._do_starttls()
                        else:
                            resp = Response(Condition.NO, text='Bad command.')
                    else:
                        if isinstance(cmd, UnauthenticateCommand):
                            resp = await self._do_unauthenticate()
                        else:
                            resp = await self._state.run(cmd)
                except Exception:
                    _log.exception('Unhandled exception')
                    resp = Response(Condition.NO, text='Server error.')
            await self._write_response(resp)
            done = resp.is_bye
        self._print('%d ---| %s', b'<disconnected>')




import unittest

from pymap.parsing import Params, ExpectedParseable, Space, EndLine
from pymap.parsing.exceptions import NotParseable
from pymap.parsing.primitives import Nil, Number, Atom, String, QuotedString, \
    List


class TestExpectedParseable(unittest.TestCase):

    def test_parse(self):
        nil, _ = ExpectedParseable.parse(b'nil', Params(expected=[Nil]))
        self.assertIsInstance(nil, Nil)
        num, _ = ExpectedParseable.parse(b'123', Params(expected=[Number]))
        self.assertIsInstance(num, Number)
        atom, _ = ExpectedParseable.parse(b'ATOM', Params(expected=[Atom]))
        self.assertIsInstance(atom, Atom)
        qstr, _ = ExpectedParseable.parse(b'"test"', Params(expected=[String]))
        self.assertIsInstance(qstr, QuotedString)
        list_, _ = ExpectedParseable.parse(b'()', Params(expected=[List]))
        self.assertIsInstance(list_, List)

    def test_parse_expectation_failure(self):
        with self.assertRaises(NotParseable):
            ExpectedParseable.parse(b'ATOM', Params(expected=[Number, Nil]))

    def test_parse_expectation_casting(self):
        num, _ = ExpectedParseable.parse(b'123', Params(expected=[Atom]))
        self.assertIsInstance(num, Atom)
        nil, _ = ExpectedParseable.parse(b'nil', Params(expected=[Atom]))
        self.assertIsInstance(num, Atom)


class TestSpace(unittest.TestCase):

    def test_parse(self):
        ret, buf = Space.parse(b'    ', Params())
        self.assertIsInstance(ret, Space)
        self.assertEqual(4, ret.length)

    def test_parse_failure(self):
        with self.assertRaises(NotParseable):
            Space.parse(b'test ', Params())
        with self.assertRaises(NotParseable):
            Space.parse(b'', Params())

    def test_bytes(self):
        ret = Space(4)
        self.assertEqual(b'    ', bytes(ret))


class TestEndLine(unittest.TestCase):

    def test_parse(self):
        ret, buf = EndLine.parse(b'  \r\n', Params())
        self.assertIsInstance(ret, EndLine)
        self.assertEqual(2, ret.preceding_spaces)
        self.assertTrue(ret.carriage_return)

    def test_parse_no_cr(self):
        ret, buf = EndLine.parse(b'  \n', Params())
        self.assertIsInstance(ret, EndLine)
        self.assertEqual(2, ret.preceding_spaces)
        self.assertFalse(ret.carriage_return)

    def test_parse_failure(self):
        with self.assertRaises(NotParseable):
            EndLine.parse(b'  \r', Params())
        with self.assertRaises(NotParseable):
            EndLine.parse(b' test \r\n', Params())

    def test_bytes(self):
        endl1 = EndLine(4, True)
        self.assertEqual(b'    \r\n', bytes(endl1))
        endl2 = EndLine(0, False)
        self.assertEqual(b'\n', bytes(endl2))




import unittest
from datetime import datetime, timezone

from pymap.parsing import Params
from pymap.parsing.exceptions import NotParseable
from pymap.parsing.command.auth import CreateCommand, AppendCommand, \
    ListCommand, RenameCommand, StatusCommand
from pymap.parsing.specials import StatusAttribute, Flag
from pymap.parsing.state import ParsingState, ParsingInterrupt, \
    ExpectContinuation


class TestCommandMailboxArg(unittest.TestCase):

    def test_parse(self):
        ret, buf = CreateCommand.parse(b' inbox\n  ', Params())
        self.assertEqual('INBOX', ret.mailbox)
        self.assertEqual(b'  ', buf)


class TestAppendCommand(unittest.TestCase):

    _epoch = datetime(1970, 1, 1, 1, 1, tzinfo=timezone.utc)
    _seen = Flag(br'\Seen')

    def test_parse(self):
        state = ParsingState(continuations=[b'test test!\n  '])
        ret, buf = AppendCommand.parse(
                b' inbox (\\Seen) "01-Jan-1970 01:01:00 +0000" {10}\n',
                Params(state))
        self.assertEqual('INBOX', ret.mailbox)
        self.assertEqual(1, len(ret.messages))
        self.assertEqual(b'test test!', ret.messages[0].literal)
        self.assertEqual({self._seen}, ret.messages[0].flag_set)
        self.assertEqual(self._epoch, ret.messages[0].when)
        self.assertEqual(b'  ', buf)

    def test_parse_multi(self):
        state = ParsingState(continuations=[b'test test! {14}\n'])
        with self.assertRaises(ParsingInterrupt) as raised:
            AppendCommand.parse(b' inbox {10}\n', Params(state))
        expected = raised.exception.expected
        self.assertIsInstance(expected, ExpectContinuation)
        self.assertEqual(14, expected.literal_length)
        state = ParsingState(continuations=[b'test test! {14}\n',
                                            b'second message\n  '])
        ret, buf = AppendCommand.parse(b' inbox {10}\n', Params(state))
        self.assertIsInstance(ret, AppendCommand)
        self.assertEqual('INBOX', ret.mailbox)
        self.assertEqual(2, len(ret.messages))
        self.assertEqual(b'test test!', ret.messages[0].literal)
        self.assertEqual(b'second message', ret.messages[1].literal)
        self.assertEqual(b'  ', buf)

    def test_parse_toobig(self):
        with self.assertRaises(NotParseable) as raised:
            AppendCommand.parse(
                b' inbox {10}\n',
                Params(max_append_len=9, command_name=b'APPEND'))
        self.assertEqual(b'[TOOBIG]', bytes(raised.exception.code))


class TestListCommand(unittest.TestCase):

    def test_parse(self):
        ret, buf = ListCommand.parse(b' one two*\n  ', Params())
        self.assertEqual('one', ret.ref_name)
        self.assertEqual('two*', ret.filter)
        self.assertEqual(b'  ', buf)

    def test_parse_filter_string(self):
        ret, buf = ListCommand.parse(b' one "two*"\n  ', Params())
        self.assertEqual('one', ret.ref_name)
        self.assertEqual('two*', ret.filter)
        self.assertEqual(b'  ', buf)


class TestRenameCommand(unittest.TestCase):

    def test_parse(self):
        ret, buf = RenameCommand.parse(b' one two\n  ', Params())
        self.assertEqual('one', ret.from_mailbox)
        self.assertEqual('two', ret.to_mailbox)
        self.assertEqual(b'  ', buf)


class TestStatusCommand(unittest.TestCase):

    def test_parse(self):
        ret, buf = StatusCommand.parse(b' mbx (MESSAGES UNSEEN)\n  ', Params())
        self.assertEqual('mbx', ret.mailbox)
        self.assertEqual(2, len(ret.status_list))
        self.assertIsInstance(ret.status_list[0], StatusAttribute)
        self.assertEqual(b'MESSAGES', ret.status_list[0].value)
        self.assertIsInstance(ret.status_list[1], StatusAttribute)
        self.assertEqual(b'UNSEEN', ret.status_list[1].value)
        self.assertEqual(b'  ', buf)

    def test_parse_error(self):
        with self.assertRaises(NotParseable):
            StatusCommand.parse(b' mbx ()\n', Params())




from __future__ import annotations

import re
from collections.abc import Iterable, Iterator, Sequence
from re import Match, Pattern
from typing import Final, Optional

from .mime import MessageHeader

__all__ = ['ThreadKey']


class ThreadKey(Iterable[str]):
    """Represents a hashable key used to link messages as members of the same
    thread.

    The thread key is composed of a single message ID from the ``Message-Id``,
    ``In-Reply-To``, or ``References`` headers, along with a normalized version
    of the ``Subject`` header. If two messages share a single thread key, they
    should be assigned the same
    :attr:`~pymap.interfaces.message.MessageInterface.thread_id`.

    Args:
        msg_id: The message ID bytestring.
        subject: The normalized subject bytestring.

    """

    _pattern = re.compile(r'<[^>]*>')
    _whitespace = re.compile(r'\s+')
    _fwd_pattern = re.compile(r'\s*fwd\s*:\s*', re.I)
    _re_pattern = re.compile(r'\s*re\s*:\s*', re.I)
    _listtag_pattern = re.compile(r'\s*\[.*?\]\s*')

    __slots__ = ['_pair', '__weakref__']

    def __init__(self, msg_id: str, subject: str) -> None:
        super().__init__()
        self._pair: Final = (msg_id, subject)

    def __eq__(self, other) -> bool:
        if isinstance(other, ThreadKey):
            return self._pair == other._pair
        return super().__eq__(other)

    def __hash__(self) -> int:
        return hash(self._pair)

    def __iter__(self) -> Iterator[str]:
        return iter(self._pair)

    @classmethod
    def _encode(cls, value: str) -> str:
        return cls._whitespace.sub('', value)

    @classmethod
    def _first_match(cls, value: str, *patterns: Pattern[str]) \
            -> Optional[Match[str]]:
        for pattern in patterns:
            match = pattern.match(value)
            if match is not None:
                return match
        return None

    @classmethod
    def _subject(cls, value: str) -> str:
        match = cls._first_match(
            value, cls._fwd_pattern, cls._re_pattern, cls._listtag_pattern)
        if match is None:
            return cls._whitespace.sub(' ', value.strip())
        else:
            return cls._subject(value[match.end(0):])

    @classmethod
    def get_all(cls, header: MessageHeader) -> Sequence[ThreadKey]:
        """Return all the thread keys from the message headers.

        Args:
            header: The message header.

        """
        ret: list[ThreadKey] = []
        message_id = header.parsed.message_id
        in_reply_to = header.parsed.in_reply_to
        references = header.parsed.references
        subject = header.parsed.subject
        subject_key = cls._subject(str(subject)) if subject else ''
        if message_id is not None:
            match = cls._pattern.search(str(message_id))
            if match is not None:
                ret.append(cls(cls._encode(match.group(0)), subject_key))
        if in_reply_to is not None:
            for match in cls._pattern.finditer(str(in_reply_to)):
                ret.append(cls(cls._encode(match.group(0)), subject_key))
        if references is not None:
            for match in cls._pattern.finditer(str(references)):
                ret.append(cls(cls._encode(match.group(0)), subject_key))
        return ret




from __future__ import annotations

from datetime import datetime, timezone
from typing import Optional

from grpclib.server import Stream
from pymap.parsing.message import AppendMessage
from pymap.parsing.specials import Flag, ExtensionOptions
from pymapadmin.grpc.admin_grpc import MailboxBase
from pymapadmin.grpc.admin_pb2 import AppendRequest, AppendResponse

from . import BaseHandler

__all__ = ['MailboxHandlers']

_AppendStream = Stream[AppendRequest, AppendResponse]


class MailboxHandlers(MailboxBase, BaseHandler):
    """The GRPC handlers, executed when an admin request is received. Each
    handler should receive a request, take action, and send the response.

    See Also:
        :class:`grpclib.server.Server`

    """

    async def Append(self, stream: _AppendStream) -> None:
        """Append a message directly to a user's mailbox.

        If the backend session defines a
        :attr:`~pymap.interfaces.session.SessionInterface.filter_set`, the
        active filter implementation will be applied to the appended message,
        such that the message may be discarded, modified, or placed into a
        specific mailbox.

        For example, using the CLI client::

            $ cat message.txt | pymap-admin append user@example.com

        See ``pymap-admin append --help`` for more options.

        Args:
            stream (:class:`~grpclib.server.Stream`): The stream for the
                request and response.

        """
        request = await stream.recv_message()
        assert request is not None
        sender = request.sender if request.HasField('sender') else ''
        recipient = request.recipient \
            if request.HasField('recipient') else request.user
        mailbox = request.mailbox if request.HasField('mailbox') else 'INBOX'
        flag_set = frozenset(Flag(flag) for flag in request.flags)
        when = datetime.fromtimestamp(request.when, timezone.utc)
        append_msg = AppendMessage(request.data, when, flag_set,
                                   ExtensionOptions.empty())
        validity: Optional[int] = None
        uid: Optional[int] = None
        async with self.catch_errors('Append') as result, \
                self.login_as(stream.metadata, request.user) as identity, \
                self.with_session(identity) as session:
            if session.filter_set is not None:
                filter_value = await session.filter_set.get_active()
                if filter_value is not None:
                    compiler = session.filter_set.compiler
                    filter_ = await compiler.compile(filter_value)
                    new_mailbox, append_msg = await filter_.apply(
                        sender, recipient, mailbox, append_msg)
                    if new_mailbox is None:
                        await stream.send_message(AppendResponse())
                        return
                    else:
                        mailbox = new_mailbox
            append_uid, _ = await session.append_messages(
                mailbox, [append_msg])
            validity = append_uid.validity
            uid = next(iter(append_uid.uids))
        resp = AppendResponse(result=result, mailbox=mailbox,
                              validity=validity, uid=uid)
        await stream.send_message(resp)




from __future__ import annotations

from abc import abstractmethod
from collections.abc import Sequence
from typing import Optional

from .interfaces.filter import FilterValueT, FilterCompilerInterface, \
    FilterSetInterface
from .plugin import Plugin

__all__ = ['filters', 'PluginFilterSet', 'SingleFilterSet']

#: Registers filter compiler plugins.
filters: Plugin[type[FilterCompilerInterface]] = Plugin(
    'pymap.filter', default='sieve')


class PluginFilterSet(FilterSetInterface[FilterValueT]):
    """Base class for filter set implementations that use a filter compiler
    declared in the ``pymap.filter`` entry point. The declared entry points
    must sub-class :class:`FilterCompiler`.

    Args:
        plugin_name: The filter plugin name, or ``None`` for default.
        value_type: The filter value representation type.

    """

    def __init__(self, plugin_name: Optional[str],
                 value_type: type[FilterValueT]) -> None:
        super().__init__()
        self._plugin_name = plugin_name
        self._value_type = value_type
        self._compiler: Optional[FilterCompilerInterface[FilterValueT]] = None

    @property
    def compiler(self) -> FilterCompilerInterface[FilterValueT]:
        if self._compiler is None:
            name = self._plugin_name
            if name is not None:
                filter_cls = filters.registered[name]
            else:
                filter_cls = filters.default
            compiler = filter_cls()
            if not issubclass(compiler.value_type, self._value_type):
                raise TypeError(f'{filter_cls} does not support '
                                f'{self._value_type}')
            self._compiler = compiler
        return self._compiler


class SingleFilterSet(FilterSetInterface[FilterValueT]):
    """Base class for a filter set that does not use named filter
    implementations, it contains only a single active filter implementation.

    """

    @property
    def name(self) -> str:
        """The permanent name for the active filter."""
        return 'active'

    async def put(self, name: str, value: FilterValueT) -> None:
        if name == self.name:
            await self.replace_active(value)

    async def delete(self, name: str) -> None:
        if name == self.name:
            await self.replace_active(None)
        else:
            raise KeyError(name)

    async def rename(self, before_name: str, after_name: str) -> None:
        raise NotImplementedError()

    async def clear_active(self) -> None:
        raise NotImplementedError()

    async def set_active(self, name: str) -> None:
        if name != self.name:
            raise KeyError(name)

    async def get(self, name: str) -> FilterValueT:
        if name == self.name:
            value = await self.get_active()
            if value is None:
                raise KeyError(name)
            else:
                return value
        else:
            raise KeyError(name)

    async def get_all(self) -> tuple[Optional[str], Sequence[str]]:
        value = await self.get_active()
        if value is None:
            return None, []
        else:
            return self.name, [self.name]

    async def replace_active(self, value: Optional[FilterValueT]) -> None:
        """Replace the current active filter value with a new value.

        Args:
            value: The new filter value.

        """
        raise NotImplementedError()

    @abstractmethod
    async def get_active(self) -> Optional[FilterValueT]:
        ...




import unittest
from collections.abc import Callable
from typing import TypeVar, Optional
from unittest.mock import MagicMock

from pymap.health import HealthStatus

_T = TypeVar('_T')


class TestHealthStatus(unittest.TestCase):

    def setUp(self) -> None:
        self._callback = MagicMock()

    def _check_call(self, healthy: Optional[bool],
                    func: Callable[..., _T], *args) -> _T:
        ret = func(*args)
        if healthy is None:
            self._callback.assert_not_called()
        else:
            self._callback.assert_called_once_with(healthy)
        self._callback.reset_mock()
        return ret

    def test_name(self) -> None:
        status = HealthStatus(name='foo')
        self.assertEqual('foo', status.name)
        child1 = status.new_dependency(name='bar')
        self.assertEqual('foo.bar', child1.name)
        child2 = status.new_dependency(name='baz')
        self.assertEqual('foo.baz', child2.name)
        child3 = child1.new_dependency(name='oof')
        self.assertEqual('foo.bar.oof', child3.name)

    def test_set(self) -> None:
        status = HealthStatus()
        self.assertTrue(status.healthy)
        status = HealthStatus(False)
        self.assertFalse(status.healthy)
        status.set(True)
        self.assertTrue(status.healthy)
        status.set_unhealthy()
        self.assertFalse(status.healthy)
        status.set_healthy()
        self.assertTrue(status.healthy)

    def test_callback(self) -> None:
        status = HealthStatus()
        self._check_call(True, status.register, self._callback)
        self._check_call(False, status.set_unhealthy)
        self._check_call(None, status.set_unhealthy)
        self._check_call(True, status.set_healthy)
        self._check_call(None, status.set_healthy)

    def test_add_dependency(self) -> None:
        parent = HealthStatus()
        self._check_call(True, parent.register, self._callback)
        deps = [self._check_call(None, parent.new_dependency, True),
                self._check_call(False, parent.new_dependency, False)]
        self._check_call(None, deps[0].set_healthy)
        self._check_call(True, deps[1].set_healthy)
        self._check_call(False, parent.set_unhealthy)
        self._check_call(True, parent.set_healthy)
        self._check_call(False, deps[0].set_unhealthy)
        self._check_call(None, deps[1].set_unhealthy)
        self._check_call(None, deps.pop)
        self._check_call(True, deps.clear)




from __future__ import annotations

import re
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from re import Pattern
from typing import Optional

__all__ = ['ListEntry', 'ListTree']


@dataclass(frozen=True)
class ListEntry:
    """An entry in the list results.

    Args:
        name: The name of the mailbox.
        exists: False if the mailbox should be marked ``\\Noselect``.
        marked: True, False, or None if the mailbox should be marked with
            ``\\Marked``, ``\\Unmarked`` , or neither, respectively.
        has_children: Whether the mailbox should be marked ``\\HasChildren`` or
            ``\\HasNoChildren``.

    """

    name: str
    exists: bool
    marked: Optional[bool]
    has_children: bool

    @property
    def attributes(self) -> Sequence[bytes]:
        """The mailbox attributes that should be returned with the mailbox
        in a ``LIST`` response, e.g. ``\\Noselect``.

        See Also:
            `RFC 3348 <https://tools.ietf.org/html/rfc3348>`_

        """
        ret: list[bytes] = []
        if not self.exists:
            ret.append(b'Noselect')
        if self.has_children:
            ret.append(b'HasChildren')
        else:
            ret.append(b'HasNoChildren')
        if self.marked is True:
            ret.append(b'Marked')
        elif self.marked is False:
            ret.append(b'Unmarked')
        return ret


class _TreeNode:

    __slots__ = ['parent', 'name', 'exists', 'children']

    def __init__(self, name: str, parent: _TreeNode = None) -> None:
        super().__init__()
        self.parent = parent
        self.name = name
        self.exists = False
        self.children: dict[str, _TreeNode] = {}

    def add(self, node_name: str, *extra: str) -> None:
        child = self.children.get(node_name)
        if not child:
            self.children[node_name] = child = _TreeNode(node_name, self)
        if not extra:
            child.exists = True
        else:
            child.add(*extra)


class ListTree:
    """Constructs a tree of hierarchical mailbox names. If a mailbox name
    has superior names in the heirarchy that do not exist, they are added as
    "unreferenced".

    Args:
        delimiter: The string delimiter for nested mailbox parts.

    """

    _wildcards = re.compile(r'([\*\%])')

    __slots__ = ['_delimiter', '_no_delimiter', '_root', '_marked']

    def __init__(self, delimiter: str) -> None:
        super().__init__()
        self._delimiter = delimiter
        self._no_delimiter = '[^' + re.escape(delimiter) + ']*?'
        self._root = _TreeNode('')
        self._marked: dict[str, bool] = {}

    def update(self, *names: str) -> ListTree:
        """Add all the mailbox names to the tree, filling in any missing nodes.

        Args:
            names: The names of the mailboxes.

        """
        for name in names:
            parts = name.split(self._delimiter)
            self._root.add(*parts)
        return self

    def set_marked(self, name: str, marked: bool = False,
                   unmarked: bool = False) -> None:
        """Add or remove the ``\\Marked`` and ``\\Unmarked`` mailbox
        attributes.

        Args:
            name: The name of the mailbox.
            marked: True if the ``\\Marked`` attribute should be added.
            unmarked: True if the ``\\Unmarked`` attribute should be added.

        """
        if marked:
            self._marked[name] = True
        elif unmarked:
            self._marked[name] = False
        else:
            self._marked.pop(name, None)

    def _iter(self, node: _TreeNode, name: str) -> Iterable[ListEntry]:
        if node.parent is not None:
            marked = self._marked.get(name)
            yield ListEntry(name, node.exists, marked, bool(node.children))
        for child in node.children.values():
            if name:
                child_name = self._delimiter.join((name, child.name))
            else:
                child_name = child.name
            for entry in self._iter(child, child_name):
                yield entry

    def _find(self, node: _TreeNode, node_name: str, *extra: str) -> _TreeNode:
        child = node.children[node_name]
        if extra:
            return self._find(child, *extra)
        else:
            return child

    def get(self, name: str) -> Optional[ListEntry]:
        """Return the named entry in the list tree.

        Args:
            name: The entry name.

        """
        parts = name.split(self._delimiter)
        try:
            node = self._find(self._root, *parts)
        except KeyError:
            return None
        else:
            marked = self._marked.get(name)
            return ListEntry(name, node.exists, marked, bool(node.children))

    def get_renames(self, from_name: str, to_name: str) \
            -> Sequence[tuple[str, str]]:
        """Return a list of tuples for all mailboxes that must be renamed, for
        the given rename operation. This should include
        ``(from_name, to_name)`` as well as all inferior names in the heirarchy
        that must also be renamed. If ``from_name`` does not exist, an empty
        list is returned.

        See Also:
            `RFC 3501 6.3.5
            <https://tools.ietf.org/html/rfc3501#section-6.3.5>`_

        Args:
            from_name: The original name of the mailbox.
            to_name: The intended new name of the mailbox.

        """
        from_parts = from_name.split(self._delimiter)
        try:
            from_node = self._find(self._root, *from_parts)
        except KeyError:
            return []
        from_names = (entry.name for entry in self._iter(from_node, from_name)
                      if entry.exists)
        to_names = (entry.name for entry in self._iter(from_node, to_name)
                    if entry.exists)
        return list(zip(from_names, to_names))

    def list(self) -> Iterable[ListEntry]:
        """Return all the entries in the list tree."""
        for entry in self._iter(self._root, ''):
            yield entry

    def _get_pattern(self, query: str) -> tuple[Pattern, Pattern]:
        pattern_parts: list[str] = []
        for part in self._wildcards.split(query):
            if part == '*':
                pattern_parts.append('.*?')
            elif part == '%':
                pattern_parts.append(self._no_delimiter)
            else:
                pattern_parts.append(re.escape(part))
        pattern = '^' + ''.join(pattern_parts) + '$'
        return re.compile(pattern), re.compile(pattern, re.IGNORECASE)

    def list_matching(self, ref_name: str, filter_: str) \
            -> Iterable[ListEntry]:
        """Return all the entries in the list tree that match the given query.

        Args:
            ref_name: Mailbox reference name.
            filter_: Mailbox name with possible wildcards.

        """
        canonical, canonical_i = self._get_pattern(ref_name + filter_)
        for entry in self.list():
            if entry.name == 'INBOX':
                if canonical_i.match('INBOX'):
                    yield entry
            elif canonical.match(entry.name):
                yield entry




import unittest
from typing import Final
from unittest.mock import create_autospec, Mock

from pymap.cluster import ListenCallback, ClusterMetadata


class _Member:

    def __init__(self, name: str) -> None:
        self.name: Final = name
        self.metadata: Final = {'name': name.encode('ascii')}

    def __hash__(self) -> int:
        return hash(self.name)


class TestClusterMetadata(unittest.TestCase):

    def test_local(self) -> None:
        metadata = ClusterMetadata()
        self.assertEqual(0, len(metadata.local))
        data = {'one': b'1', 'two': b'2', 'three': b'3'}
        metadata.local.update(data)
        self.assertEqual(data, metadata.local)
        del metadata.local['two']
        self.assertEqual(b'1', metadata.local['one'])
        self.assertNotIn('two', metadata.local)
        self.assertEqual(b'3', metadata.local['three'])

    def test_listen(self) -> None:
        callback = create_autospec(ListenCallback)
        arg = Mock()
        metadata = ClusterMetadata()
        metadata.local['one'] = b'1'
        metadata.listen(callback, arg)
        callback.assert_called_with(arg, {'one': b'1'})
        metadata.local['two'] = b'2'
        callback.assert_called_with(arg, {'one': b'1', 'two': b'2'})
        del metadata.local['two']
        callback.assert_called_with(arg, {'one': b'1'})
        metadata.local['one'] = b'one'
        callback.assert_called_with(arg, {'one': b'one'})
        metadata.local
        callback.reset_mock()
        del arg
        metadata.local['three'] = b'3'
        callback.assert_not_called()
        self.assertEqual({'one': b'one', 'three': b'3'}, metadata.local)

    def test_remote(self) -> None:
        member1 = _Member('one')
        member2 = _Member('two')
        metadata = ClusterMetadata([member1, member2])
        self.assertEqual({member1, member2}, set(metadata))
        self.assertEqual(2, len(metadata))
        self.assertIn(member1, metadata)
        self.assertIn(member2, metadata)
        self.assertEqual(1, len(metadata.remote))
        self.assertEqual({'name': {member1: b'one', member2: b'two'}},
                         metadata.remote)
        del member2
        self.assertEqual({'name': {member1: b'one'}}, metadata.remote)
        metadata.discard(member1)
        self.assertEqual({'name': {}}, metadata.remote)

    def test_get_all(self) -> None:
        member1 = _Member('one')
        member2 = _Member('two')
        metadata = ClusterMetadata([member1, member2])
        metadata.local['name'] = b'three'
        self.assertEqual({b'one', b'two', b'three'}, metadata.get_all('name'))




from __future__ import annotations

import json
import logging
import uuid
from argparse import ArgumentParser, Namespace
from asyncio import CancelledError
from collections.abc import Awaitable, Callable, Mapping, AsyncIterator
from contextlib import asynccontextmanager, suppress, AsyncExitStack
from datetime import datetime
from secrets import token_bytes
from typing import Any, Optional, Final

from aioredis import Redis, ConnectionError
from pysasl.creds import AuthenticationCredentials

from pymap.bytes import BytesFormat
from pymap.config import BackendCapability, IMAPConfig
from pymap.context import connection_exit
from pymap.exceptions import AuthorizationFailure, IncompatibleData, \
    NotAllowedError, UserNotFound
from pymap.health import HealthStatus
from pymap.interfaces.backend import BackendInterface
from pymap.interfaces.login import LoginInterface, IdentityInterface
from pymap.interfaces.token import TokensInterface
from pymap.token import AllTokens
from pymap.user import UserMetadata

from .background import BackgroundAction, BackgroundTask, NoopAction
from .cleanup import CleanupAction
from .filter import FilterSet
from .keys import DATA_VERSION, RedisKey, GlobalKeys, CleanupKeys, \
    NamespaceKeys
from .mailbox import Message, MailboxSet
from ..session import BaseSession

__all__ = ['RedisBackend', 'Config', 'Session']

_log = logging.getLogger(__name__)

_Connect = Callable[[], Awaitable[Redis]]


class RedisBackend(BackendInterface):
    """Defines a backend that uses redis data structures for mailbox storage.

    """

    def __init__(self, login: Login, config: Config,
                 status: HealthStatus) -> None:
        super().__init__()
        self._login = login
        self._config = config
        self._status = status

    @property
    def login(self) -> Login:
        return self._login

    @property
    def config(self) -> Config:
        return self._config

    @property
    def status(self) -> HealthStatus:
        return self._status

    @classmethod
    def add_subparser(cls, name: str, subparsers: Any) -> ArgumentParser:
        parser = subparsers.add_parser(name, help='redis backend')
        parser.add_argument('--address', metavar='URL',
                            default='redis://localhost',
                            help='the redis server address')
        parser.add_argument('--data-address', metavar='URL',
                            help='the redis server address for mail data,'
                            ' if different')
        parser.add_argument('--separator', metavar='CHAR', default='/',
                            help='the redis key segment separator')
        parser.add_argument('--prefix', metavar='VAL', default='/mail',
                            help='the mail data key prefix')
        parser.add_argument('--users-prefix', metavar='VAL', default='/users',
                            help='the user lookup key prefix')
        parser.add_argument('--users-json', action='store_true',
                            help='the user lookup value contains JSON')
        return parser

    @classmethod
    async def init(cls, args: Namespace, **overrides: Any) \
            -> tuple[RedisBackend, Config]:
        config = Config.from_args(args)
        status = HealthStatus(name='redis')
        login = Login(config, status)
        return cls(login, config, status), config

    async def start(self, stack: AsyncExitStack) -> None:
        config = self._config
        global_keys = config._global_keys
        self.status.set_healthy()
        user_action = NoopAction()
        mail_action = CleanupAction(global_keys)
        self.login._start_background(stack, user_action, mail_action)


class Config(IMAPConfig):
    """The config implementation for the redis backend.

    Args:
        args: The command-line arguments.
        address: The redis server address.
        data_address: The redis server address for mail data, if different.
        separator: The redis key segment separator.
        prefix: The prefix for mail data keys.
        users_prefix: The user lookup key prefix.
        users_json: True if the user lookup value contains JSON.

    """

    def __init__(self, args: Namespace, *, address: str,
                 data_address: Optional[str],
                 separator: bytes, prefix: bytes, users_prefix: bytes,
                 users_json: bool, **extra: Any) -> None:
        super().__init__(args, admin_key=token_bytes(), **extra)
        self._address = address
        self._data_address = data_address
        self._separator = separator
        self._prefix = prefix
        self._users_prefix = users_prefix
        self._users_json = users_json

    @property
    def backend_capability(self) -> BackendCapability:
        return BackendCapability(idle=True, object_id=True, multi_append=True)

    @property
    def address(self) -> str:
        """The redis server address for user data. Defaults to a connection to
        localhost.

        See Also:
            :func:`aioredis.create_connection`

        """
        return self._address

    @property
    def data_address(self) -> str:
        """The redis server address for mail data. Defaults to a connection to
        localhost.

        See Also:
            :func:`aioredis.create_connection`

        """
        if self._data_address is not None:
            return self._data_address
        else:
            return self.address

    @property
    def separator(self) -> bytes:
        """The bytestring used to separate segments of composite redis keys."""
        return self._separator

    @property
    def prefix(self) -> bytes:
        """The prefix for mail data keys. This prefix does not apply to
        :attr:`.users_key`.

        """
        return self._prefix

    @property
    def users_prefix(self) -> bytes:
        """The prefix for user lookup keys."""
        return self._users_prefix

    @property
    def users_json(self) -> bool:
        """True if the value from the user lookup key contains a JSON object
        with a ``"password"`` attribute, instead of a redis hash with a
        ``password`` key.

        See Also:
            `redis hashes
            <https://redis.io/topics/data-types-intro#redis-hashes>`_

        """
        return self._users_json

    @property
    def _joiner(self) -> BytesFormat:
        return BytesFormat(self.separator)

    @property
    def _users_root(self) -> RedisKey:
        return RedisKey(self._joiner, [self.users_prefix], {})

    @property
    def _global_keys(self) -> GlobalKeys:
        key = RedisKey(self._joiner, [self.prefix], {})
        return GlobalKeys(key)

    @classmethod
    def parse_args(cls, args: Namespace) -> Mapping[str, Any]:
        return {**super().parse_args(args),
                'address': args.address,
                'data_address': args.data_address,
                'separator': args.separator.encode('utf-8'),
                'prefix': args.prefix.encode('utf-8'),
                'users_prefix': args.users_prefix.encode('utf-8'),
                'users_json': args.users_json}


class Session(BaseSession[Message]):
    """The session implementation for the redis backend."""

    resource = __name__

    def __init__(self, redis: Redis, owner: str, config: Config,
                 mailbox_set: MailboxSet, filter_set: FilterSet) -> None:
        super().__init__(owner)
        self._redis = redis
        self._config = config
        self._mailbox_set = mailbox_set
        self._filter_set = filter_set

    @property
    def config(self) -> IMAPConfig:
        return self._config

    @property
    def mailbox_set(self) -> MailboxSet:
        return self._mailbox_set

    @property
    def filter_set(self) -> FilterSet:
        return self._filter_set


class Login(LoginInterface):
    """The login implementation for the redis backend."""

    def __init__(self, config: Config, status: HealthStatus) -> None:
        super().__init__()
        self._config = config
        self._tokens = AllTokens()
        self._user_redis = Redis.from_url(config.address)
        self._user_status = status.new_dependency(False, name='user')
        self._mail_redis = Redis.from_url(config.data_address)
        self._mail_status = status.new_dependency(False, name='mail')

    @property
    def tokens(self) -> TokensInterface:
        return self._tokens

    @classmethod
    async def _connect(cls, stack: AsyncExitStack,
                       redis: Redis, status: HealthStatus) -> Redis:
        try:
            conn = await stack.enter_async_context(redis.client())
        except (ConnectionError, OSError) as exc:
            is_debug = _log.isEnabledFor(logging.DEBUG)
            _log.warn('%s: %s', type(exc).__name__, exc, exc_info=is_debug)
            status.set_unhealthy()
            raise CancelledError() from exc
        else:
            status.set_healthy()
            return conn

    async def _user_connect(self) -> Redis:
        return await self._connect(connection_exit.get(),
                                   self._user_redis, self._user_status)

    async def _mail_connect(self) -> Redis:
        return await self._connect(connection_exit.get(),
                                   self._mail_redis, self._mail_status)

    def _start_background(self, stack: AsyncExitStack,
                          user_action: BackgroundAction,
                          mail_action: BackgroundAction) -> None:
        user_background = BackgroundTask(self._user_redis, self._user_status,
                                         user_action)
        mail_background = BackgroundTask(self._mail_redis, self._mail_status,
                                         mail_action)
        user_task = user_background.start()
        mail_task = mail_background.start()
        stack.callback(user_task.cancel)
        stack.callback(mail_task.cancel)

    async def _check(self) -> None:
        async with AsyncExitStack() as stack:
            with suppress(Exception):
                await self._connect(stack, self._user_redis, self._user_status)
            with suppress(Exception):
                await self._connect(stack, self._mail_redis, self._mail_status)

    async def authenticate(self, credentials: AuthenticationCredentials) \
            -> Identity:
        config = self._config
        authcid = credentials.authcid
        token_key: Optional[bytes] = None
        role: Optional[str] = None
        if credentials.authcid_type == 'admin-token':
            authcid = credentials.identity
            role = 'admin'
        try:
            authcid_identity = Identity(config, self.tokens,
                                        self._user_connect, self._mail_connect,
                                        authcid)
            metadata = await authcid_identity.get()
        except UserNotFound:
            metadata = UserMetadata(config)
        if 'key' in metadata.params:
            token_key = bytes.fromhex(metadata.params['key'])
        role = role or metadata.role
        await metadata.check_password(credentials, token_key=token_key)
        if role != 'admin' and authcid != credentials.identity:
            raise AuthorizationFailure()
        return Identity(config, self.tokens,
                        self._user_connect, self._mail_connect,
                        credentials.identity, role)


class Identity(IdentityInterface):
    """The identity implementation for the redis backend."""

    def __init__(self, config: Config, tokens: TokensInterface,
                 user_connect: _Connect, mail_connect: _Connect,
                 name: str, role: str = None) -> None:
        super().__init__()
        self.config: Final = config
        self.tokens: Final = tokens
        self._user_connect = user_connect
        self._mail_connect = mail_connect
        self._name = name
        self._role = role

    @property
    def name(self) -> str:
        return self._name

    async def new_token(self, *, expiration: datetime = None) -> Optional[str]:
        metadata = await self.get()
        if 'key' not in metadata.params:
            return None
        key = bytes.fromhex(metadata.params['key'])
        return self.tokens.get_login_token(self.name, key)

    @asynccontextmanager
    async def new_session(self) -> AsyncIterator[Session]:
        config = self.config
        conn = await self._mail_connect()
        global_keys = config._global_keys
        namespace = await self._get_namespace(conn, global_keys, self.name)
        ns_keys = NamespaceKeys(global_keys, namespace)
        cl_keys = CleanupKeys(global_keys)
        mailbox_set = MailboxSet(conn, ns_keys, cl_keys)
        filter_set = FilterSet(conn, ns_keys)
        try:
            await mailbox_set.add_mailbox('INBOX')
        except ValueError:
            pass
        yield Session(conn, self.name, config, mailbox_set, filter_set)

    async def _get_namespace(self, conn: Redis, global_keys: GlobalKeys,
                             user: str) -> bytes:
        user_key = user.encode('utf-8')
        new_namespace = uuid.uuid4().hex.encode('ascii')
        ns_val = b'%d/%b' % (DATA_VERSION, new_namespace)
        async with conn.pipeline() as multi:
            multi.hsetnx(global_keys.namespaces, user_key, ns_val)
            multi.hget(global_keys.namespaces, user_key)
            _, ns_val = await multi.execute()
        version, namespace = ns_val.split(b'/', 1)
        if int(version) != DATA_VERSION:
            raise IncompatibleData()
        return namespace

    async def get(self) -> UserMetadata:
        conn = await self._user_connect()
        user_bytes = self.name.encode('utf-8')
        user_key = self.config._users_root.end(user_bytes)
        if self.config.users_json:
            json_data = await conn.get(user_key)
            if json_data is None:
                raise UserNotFound(self.name)
            data_dict = json.loads(json_data)
        else:
            data_dict = await conn.hgetall(user_key)
            if data_dict is None:
                raise UserNotFound(self.name)
        return UserMetadata(self.config, **data_dict)

    async def set(self, metadata: UserMetadata) -> None:
        config = self.config
        conn = await self._user_connect()
        if self._role != 'admin' and metadata.role:
            raise NotAllowedError('Cannot assign role.')
        user_key = config._users_root.end(self.name.encode('utf-8'))
        user_dict = metadata.to_dict(key=token_bytes().hex())
        if self.config.users_json:
            json_data = json.dumps(user_dict)
            await conn.set(user_key, json_data)
        else:
            async with conn.pipeline() as multi:
                multi.delete(user_key)
                multi.hmset(user_key, user_dict)
                await multi.execute()

    async def delete(self) -> None:
        config = self.config
        conn = await self._user_connect()
        user_key = config._users_root.end(self.name.encode('utf-8'))
        if not await conn.delete(user_key):
            raise UserNotFound(self.name)




from __future__ import annotations

import hashlib
import os.path
from collections.abc import Sequence
from contextlib import closing
from typing import Generic, TypeVar, Any, Union, Optional

import msgpack
from aioredis import Redis
from aioredis.exceptions import NoScriptError
from pkg_resources import resource_stream

__all__ = ['ScriptBase']

_Val = Union[int, bytes]
_RetT = TypeVar('_RetT')


class ScriptBase(Generic[_RetT]):
    """The base class for redis scripts.

    Args:
        name: The script name, which should correspond to a ``.lua`` file.

    """

    __slots__ = ['_name', '_sha']

    def __init__(self, name: str) -> None:
        super().__init__()
        self._name = name
        self._sha, _ = self._load()

    def _load(self) -> tuple[str, bytes]:
        fname = os.path.join('lua', f'{self._name}.lua')
        with closing(resource_stream(__name__, fname)) as script:
            data = script.read()
        return hashlib.sha1(data).hexdigest(), data

    def _convert(self, ret: Any) -> _RetT:
        return ret

    def _pack(self, val: Any) -> bytes:
        return msgpack.packb(val)

    def _maybe_int(self, val: bytes) -> Optional[int]:
        if val == b'':
            return None
        else:
            return int(val)

    async def eval(self, redis: Redis, keys: Sequence[bytes],
                   args: Sequence[_Val]) -> _RetT:
        """Run the script.

        Notes:
            This attempts the EVALSHA command first, then falls back to EVAL if
            a NOSCRIPT error is returned.

        Raises:
            :class:`~asyncio.ResponseError`

        """
        num_keys = len(keys)
        # TODO: remove after aioredis > 2.0.0
        tmp_args: list[str] = [*keys, *args]  # type: ignore
        try:
            ret = await redis.evalsha(self._sha, num_keys, *tmp_args)
        except NoScriptError:
            self._sha, data = self._load()
            ret = await redis.eval(data, num_keys, *keys, *args)
        return self._convert(ret)




from __future__ import annotations

import base64
from collections.abc import Iterable, Mapping, Sequence
from email.headerregistry import ContentTypeHeader
from email.policy import SMTP
from itertools import chain, islice
from typing import Any, Optional, Final

from .parsed import ParsedHeaders
from ._util import whitespace, find_any, get_raw
from ..bytes import WriteStream, Writeable

__all__ = ['MessageContent', 'MessageHeader', 'MessageBody']

_default_type: Final = 'text/plain'

_Line = tuple[int, int, int]
_Lines = Sequence[_Line]
_Folded = Sequence[tuple[str, _Lines]]


class MessageContent(Writeable):
    """Contains the message content, parsed for IMAP processing.

    Args:
        data: The message literal.
        header: The parsed message header.
        body: The parsed message body.

    Attributes:
        lines: The number of lines in the message content.

    """

    __slots__ = ['_raw', 'lines', 'header', 'body', '__weakref__']

    def __init__(self, data: bytes, header: MessageHeader,
                 body: MessageBody) -> None:
        super().__init__()
        self._raw = get_raw(memoryview(data), header._lines, body._lines)
        self.lines: Final = header.lines + body.lines - 1
        self.header: Final = header
        self.body: Final = body

    def walk(self) -> Iterable[MessageContent]:
        """Iterate through the message and all its nested sub-parts in the
        order they occur.

        """
        if self.body.has_nested:
            return chain([self], *(part.walk() for part in self.body.nested))
        else:
            return [self]

    @property
    def is_rfc822(self) -> bool:
        """True if the content-type of the message is ``message/rfc822``."""
        ct_hdr = self.header.parsed.content_type
        if ct_hdr is None:
            return False
        else:
            return ct_hdr.content_type == 'message/rfc822'

    @property
    def json(self) -> Mapping[str, Any]:
        """A dictionary that can be serialized (e.g. with :mod:`json`), so that
        this object may be re-created without parsing.

        See Also:
            :meth:`.from_json`

        """
        return {'header': self.header.json,
                'body': self.body.json}

    @classmethod
    def from_json(cls, data: bytes, json: Mapping[str, Any]) -> MessageContent:
        """Recover the parsed message content without re-parsing, using the
        original raw data and the :attr:`.json`.

        In this example, ``content1`` and ``content2`` should be equivalent::

            message = b'...'
            content1 = MessageContent.parse(message)
            content2 = MessageContent.from_json(message, content1.json)

        Args:
            data: The bytestring that was parsed.
            json: The :attr:`.json` of a previously parsed message content.

        """
        header = MessageHeader.from_json(data, json['header'])
        body = MessageBody.from_json(data, json['body'])
        return cls(data, header, body)

    @classmethod
    def parse(cls, data: bytes) -> MessageContent:
        """Parse the bytestring into message content.

        Args:
            data: The bytestring to parse.

        """
        lines = cls._find_lines(data)
        view = memoryview(data)
        return cls._parse(data, view, lines)

    @classmethod
    def _parse(cls, data: bytes, view: memoryview, lines: _Lines) \
            -> MessageContent:
        header_lines, body_lines = cls._split_lines(data, lines)
        header = MessageHeader._parse(data, view, header_lines)
        content_type = header.parsed.content_type
        body = MessageBody._parse(data, view, body_lines, content_type)
        return cls(data, header, body)

    @classmethod
    def _find_lines(cls, data: bytes) -> _Lines:
        start = 0
        end = len(data)
        ret: list[_Line] = []
        while True:
            idx = data.find(b'\n', start, end)
            if idx < 0:
                ret.append((start, end, end))
                break
            next_start = idx + 1
            idx_minus_1 = idx - 1
            if idx_minus_1 >= start and data[idx_minus_1] == 0x0d:
                idx = idx_minus_1
            ret.append((start, idx, next_start))
            start = next_start
        return ret

    @classmethod
    def _split_lines(cls, data: bytes, lines: _Lines) -> tuple[_Lines, _Lines]:
        for i, line in enumerate(lines):
            start, end, _ = line
            ws_end = find_any(data, whitespace, start, end, False, False)
            if ws_end < 0:
                return lines[0:i + 1], lines[i + 1:]
        return [], lines

    def write(self, writer: WriteStream) -> None:
        writer.write(bytes(self))

    def __len__(self) -> int:
        return len(self._raw)

    def __bytes__(self) -> bytes:
        return bytes(self._raw)


class MessageHeader(Writeable):
    """The message header. Contains lines in the form of ``Header: value\\n``,
    possibly folded onto multiple lines where subsequent lines start with
    whitespace.

    Attributes:
        folded: A list of headers, as they occurred in the original data, as
            tuples of the lower-cased header name and the full header value,
            including the header name and any extra folded lines.
        parsed: The message headers, as a dictionary-like object that parses
            headers on-demand.

    """

    __slots__ = ['_raw', '_lines', '_folded', 'folded', 'parsed']

    def __init__(self, data: bytes, lines: _Lines, folded: _Folded) -> None:
        super().__init__()
        view = memoryview(data)
        self._raw = get_raw(view, lines)
        self._lines = lines
        self._folded = folded
        self.folded: Final = self._get_folded(view, folded)
        self.parsed: Final = self._get_parsed(data, folded)

    @property
    def lines(self) -> int:
        """The number of lines in the message header."""
        return len(self._lines)

    @property
    def json(self) -> Mapping[str, Any]:
        """A dictionary that can be serialized (e.g. with :mod:`json`), so that
        this object may be re-created without parsing.

        See Also:
            :meth:`.from_json`

        """
        return {'lines': self._lines,
                'folded': self._folded}

    @classmethod
    def from_json(cls, data: bytes, json: Mapping[str, Any]) -> MessageHeader:
        """Recover the parsed message header without re-parsing, using the
        original raw data and the :attr:`.json`.

        See Also:
            :meth:`MessageContent.from_json`

        Args:
            data: The bytestring that was parsed.
            json: The :attr:`.json` of a previously parsed message content.

        """
        lines: _Lines = json['lines']
        folded: _Folded = json['folded']
        return cls(data, lines, folded)

    @classmethod
    def empty(cls) -> MessageHeader:
        """Return an empty header object."""
        return cls(b'', [], [])

    @classmethod
    def _get_folded(cls, view: memoryview, folded: _Folded) \
            -> Sequence[tuple[bytes, memoryview]]:
        return [(cls._to_bytes(key), get_raw(view, lines))
                for key, lines in folded]

    @classmethod
    def _get_parsed(cls, data: bytes, folded: _Folded) -> ParsedHeaders:
        header_map: dict[bytes, list[list[bytes]]] = {}
        for key, lines in folded:
            name = cls._to_bytes(key)
            values = header_map.setdefault(name, [])
            values.append([data[start:end] for start, end, _ in lines])
        return ParsedHeaders(header_map)

    @classmethod
    def _to_bytes(cls, key: str) -> bytes:
        return base64.b64decode(key)

    @classmethod
    def _to_str(cls, key: bytes) -> str:
        return base64.b64encode(key).decode('ascii')

    @classmethod
    def _parse(cls, data: bytes, view: memoryview,
               lines: _Lines) -> MessageHeader:
        folds = cls._find_folds(data, lines)
        folded = cls._find_folded(data, view, folds)
        return cls(data, lines, folded)

    @classmethod
    def _find_folds(cls, data: bytes, lines: _Lines) -> Sequence[_Lines]:
        ret: list[list[tuple[int, int, int]]] = []
        if not lines:
            return []
        for line in islice(lines, len(lines) - 1):
            start, end, _ = line
            length = end - start
            if length >= 1 and data[start] in whitespace:
                if ret:
                    ret[-1].append(line)
            else:
                ret.append([line])
        return ret

    @classmethod
    def _find_folded(cls, data: bytes, view: memoryview,
                     folds: Sequence[_Lines]) -> _Folded:
        folded: list[tuple[str, _Lines]] = []
        for group in folds:
            start, end, _ = group[0]
            colon = data.find(b':', start, end)
            if colon < 0:
                continue
            name = data[start:colon].strip().lower()
            folded.append((cls._to_str(name), group))
        return folded

    def write(self, writer: WriteStream) -> None:
        writer.write(bytes(self))

    def __len__(self) -> int:
        return len(self._raw)

    def __bytes__(self) -> bytes:
        return bytes(self._raw)


class MessageBody(Writeable):
    """The message body, starting immediately after the header. The body may
    contain nested sub-parts, which are each valid :class:`MessageContent`
    objects.

    Attributes:
        content_type: The content type of the message body.

    """

    __slots__ = ['_raw', '_lines', '_nested', 'content_type']

    def __init__(self, data: bytes, lines: _Lines,
                 content_type: ContentTypeHeader,
                 nested: Sequence[MessageContent]) -> None:
        super().__init__()
        self._raw = get_raw(memoryview(data), lines)
        self._lines = lines
        self._nested = nested
        self.content_type: Final = content_type

    @property
    def lines(self) -> int:
        """The number of lines in the message body."""
        return len(self._lines)

    @property
    def has_nested(self) -> bool:
        """True if the message body is composed of nested sub-parts."""
        return len(self.nested) > 0

    @property
    def nested(self) -> Sequence[MessageContent]:
        """If :attr:`.has_nested` is True, contains the list of sub-parts."""
        return self._nested

    @property
    def json(self) -> Mapping[str, Any]:
        """A dictionary that can be serialized (e.g. with :mod:`json`), so that
        this object may be re-created without parsing.

        See Also:
            :meth:`.from_json`

        """
        return {'lines': self._lines,
                'content_type': str(self.content_type),
                'nested': [part.json for part in self._nested]}

    @classmethod
    def from_json(cls, data: bytes, json: Mapping[str, Any]) -> MessageBody:
        """Recover the parsed message body without re-parsing, using the
        original raw data and the :attr:`.json`.

        See Also:
            :meth:`MessageContent.from_json`

        Args:
            data: The bytestring that was parsed.
            json: The :attr:`.json` of a previously parsed message content.

        """
        lines: _Lines = json['lines']
        content_type = cls._parse_content_type(json['content_type'])
        nested = [MessageContent.from_json(data, part)
                  for part in json['nested']]
        return cls(data, lines, content_type, nested)

    @classmethod
    def empty(cls) -> MessageBody:
        """Return an empty body object."""
        content_type = cls._parse_content_type(_default_type)
        return cls(b'', [], content_type, [])

    @classmethod
    def _parse(cls, data: bytes, view: memoryview, lines: _Lines,
               content_type: Optional[ContentTypeHeader]) -> MessageBody:
        if content_type is None:
            content_type = cls._parse_content_type(_default_type)
        maintype = content_type.maintype
        if maintype == 'multipart':
            boundary = cls._get_boundary(content_type)
            if boundary:
                return cls._parse_multipart(
                    data, view, lines, content_type, boundary)
        elif maintype == 'message' and content_type.subtype == 'rfc822':
            return cls._parse_rfc822(data, view, lines, content_type)
        return cls(data, lines, content_type, [])

    @classmethod
    def _parse_content_type(cls, header: str) -> ContentTypeHeader:
        val = SMTP.header_fetch_parse('Content-Type', header)
        assert isinstance(val, ContentTypeHeader)
        return val

    @classmethod
    def _get_boundary(cls, content_type: ContentTypeHeader) -> Optional[bytes]:
        try:
            boundary = content_type.params['boundary']
        except KeyError:
            pass
        else:
            if boundary:
                try:
                    return boundary.encode('ascii')
                except UnicodeError:
                    pass
        return None

    @classmethod
    def _parse_rfc822(cls, data: bytes, view: memoryview, lines: _Lines,
                      content_type: ContentTypeHeader) -> MessageBody:
        subpart = MessageContent._parse(data, view, lines)
        return cls(data, lines, content_type, [subpart])

    @classmethod
    def _parse_multipart(cls, data: bytes, view: memoryview, lines: _Lines,
                         content_type: ContentTypeHeader,
                         boundary: bytes) -> MessageBody:
        parts = cls._find_parts(data, view, lines, boundary)
        nested: list[MessageContent] = []
        for part_lines in parts:
            sub_content = MessageContent._parse(data, view, part_lines)
            nested.append(sub_content)
        return cls(data, lines, content_type, nested)

    @classmethod
    def _find_parts(cls, data: bytes, view: memoryview, lines: _Lines,
                    boundary: bytes) -> Sequence[_Lines]:
        ret: list[list[_Line]] = []
        part_match = (b'--%s' % boundary, b'--%s' % boundary)
        stop_match = (b'--%s--' % boundary, b'--%s--' % boundary)
        for line in lines:
            start, end, _ = line
            line_view = view[start:end]
            if any(line_view == m for m in stop_match):
                break
            elif any(line_view == m for m in part_match):
                ret.append([])
            elif ret:
                ret[-1].append(line)
        return ret

    def write(self, writer: WriteStream) -> None:
        writer.write(bytes(self))

    def __len__(self) -> int:
        return len(self._raw)

    def __bytes__(self) -> bytes:
        return bytes(self._raw)
